# Copyright 2015 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.
"""Module for baked spitfire utility functions."""
try:
    from spitfire.runtime import _baked  # pylint: disable=g-import-not-at-top
except ImportError:
    _baked = None


class _SanitizedPlaceholder(str):
    """Class that represents an already sanitized string.

    SanitizedPlaceholder wraps another value to let the runtime know that this
    does not need to filtered again.
    """

    def __add__(self, other):
        value = str.__add__(self, other)
        if type(other) == _SanitizedPlaceholder:
            value = _SanitizedPlaceholder(value)
        return value

    def __mod__(self, other):
        value = str.__mod__(self, other)
        if type(other) == _SanitizedPlaceholder:
            value = _SanitizedPlaceholder(value)
        return value


def _runtime_mark_as_sanitized(value, function):
    """Wrap a function's return value in a SanitizedPlaceholder.

    This function is called often so it needs to be fast. This
    function checks the skip_filter annotation on the function passed
    in to determine if the value should be wrapped.

    Args:
      value: The value to be marked
      function: The function to check for skip_filter

    Returns:
      Either a SanitizedPlaceholder object or the value passed in.
    """
    if getattr(function, 'skip_filter', False):
        if type(value) == str:
            return _SanitizedPlaceholder(value)
    return value


def _mark_as_sanitized(value):
    """Wrap a value in a SanitizedPlaceholder.

    This function is called often so it needs to be fast.

    Args:
      value: The value to be marked

    Returns:
      Either a SanitizedPlaceholder object or the value passed in.
    """
    # The if branch is going to be taken in most cases.
    if type(value) == str:
        return _SanitizedPlaceholder(value)
    return value


if _baked is None:
    SanitizedPlaceholder = _SanitizedPlaceholder
    runtime_mark_as_sanitized = _runtime_mark_as_sanitized
    mark_as_sanitized = _mark_as_sanitized
else:
    # Use C.
    SanitizedPlaceholder = _baked._SanitizedPlaceholder
    runtime_mark_as_sanitized = _baked._runtime_mark_as_sanitized
    mark_as_sanitized = _baked._mark_as_sanitized



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from spitfire.compiler import parser
from third_party.yapps2 import yappsrt

# SpitfireScanner uses the order of the match, not the length of the match to
# determine what token to return. I'm not sure how fragille this is long-term,
# but it seems to have been the right solution for a number of small problems
# allong the way.
_restrict_cache = {}


class SpitfireScanner(parser._SpitfireParserScanner):

    def token(self, i, restrict=0):
        """Get the i'th token, and if i is one past the end, then scan
        for another token; restrict is a list of tokens that
        are allowed, or 0 for any token."""
        if i == len(self.tokens):
            self.scan(restrict)
        if i < len(self.tokens):
            # Make sure the restriction is more restricted
            restriction = self.restrictions[i]
            if restrict and restriction:
                if not restriction.issuperset(restrict):
                    raise NotImplementedError(
                        "Unimplemented: restriction set changed", restrict,
                        self.restrictions[i])
                return self.tokens[i]
            elif not restrict and not restriction:
                return self.tokens[i]
        raise yappsrt.NoMoreTokens(i, len(self.tokens), self.tokens[i],
                                   restrict, self.restrictions[i], self.tokens)

    def scan(self, restrict):
        """Should scan another token and add it to the list, self.tokens,
        and add the restriction to self.restrictions"""
        # Cache the list of patterns we check to avoid unnecessary iteration
        restrict = frozenset(restrict)
        try:
            patterns = _restrict_cache[restrict]
        except KeyError:
            patterns = [pair
                        for pair in self.patterns
                        if not restrict or pair[0] in restrict]
            _restrict_cache[restrict] = patterns

        _input, _pos = self.input, self.pos
        for best_pat, regexp in patterns:
            m = regexp.match(_input, _pos)
            if m:
                tname = m.group(0)
                best_match = len(tname)
                # msolo: use the first match, not the 'best'
                break
        else:
            # If we didn't find anything, raise an error
            msg = "Bad Token"
            if restrict:
                msg = "Trying to find one of " + ', '.join(restrict)
            raise yappsrt.SyntaxError(self.pos, msg)

        # Create a token with this data
        end = _pos + best_match
        token = (_pos, end, best_pat, tname)
        self.pos = end
        # Only add this token if it's not in the list
        # (to prevent looping)
        if not self.tokens or token != self.tokens[-1]:
            self.tokens.append(token)
            self.restrictions.append(restrict)
        return




import spitfire


env = spitfire.Environment('templates')
TABLE_DATA = [
    dict(a=1,
         b=2,
         c=3,
         d=4,
         e=5)
]

print(env.render('ample.spf', 
			[{"a":1,"b":2} for x in range(4)], 
			template_name="ample"
))



# Copyright 2014 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import unittest

from spitfire.compiler import analyzer
from spitfire.compiler import ast
from spitfire.compiler import compiler
from spitfire.compiler import optimizer
from spitfire.compiler import options
from spitfire.compiler import util
from spitfire.compiler import walker
from spitfire import test_util


class BaseTest(unittest.TestCase):

    def __init__(self, *args):
        unittest.TestCase.__init__(self, *args)
        self.analyzer_options = options.default_options

    def setUp(self):
        self.compiler = compiler.Compiler(
            analyzer_options=self.analyzer_options,
            xspt_mode=False,
            compiler_stack_traces=True)

    def _get_analyzer(self, ast_root):
        optimization_analyzer = optimizer.OptimizationAnalyzer(
            ast_root, self.compiler.analyzer_options, self.compiler)
        optimization_analyzer.visit_ast = test_util.RecordedFunction(
            optimization_analyzer.visit_ast)
        return optimization_analyzer

    def _build_function_template(self):
        """ Build a simple template with a function.

        file: TestTemplate
        #def test_function
        #end def
        """
        ast_root = ast.TemplateNode('TestTemplate')
        function_node = ast.FunctionNode('test_function')
        ast_root.append(function_node)
        return (ast_root, function_node)

    def _build_if_template(self, condition=None):
        """ Build a simple template with a function and an if statement.

        file: TestTemplate
        #def test_function
          #if True
          #end if
        #end def
        """
        ast_root, function_node = self._build_function_template()
        condition_node = condition or ast.LiteralNode(True)
        if_node = ast.IfNode(condition_node)
        function_node.append(if_node)
        return (ast_root, function_node, if_node)

    def _compile(self, template_content):
        template_node = util.parse_template(template_content)
        template_node.source_path = 'test_template.spt'
        return template_node


class TestAnalyzeListLiteralNode(BaseTest):

    def test_list_elements_are_optimized(self):
        self.ast_description = """
        Input:
        [1, 2, 3]
        """

        ast_root = ast.ListLiteralNode('list')
        ast_root.child_nodes.append(ast.LiteralNode(1))
        ast_root.child_nodes.append(ast.LiteralNode(2))
        ast_root.child_nodes.append(ast.LiteralNode(3))

        optimization_analyzer = self._get_analyzer(ast_root)
        optimization_analyzer.visit_ast(ast_root)
        self.assertEqual(len(optimization_analyzer.visit_ast.GetCalls()), 4)


class TestAssignAfterFilterWarning(unittest.TestCase):

    def setUp(self):
        analyzer_options = options.default_options
        analyzer_options.update(cache_resolved_placeholders=True,
                                enable_warnings=True,
                                warnings_as_errors=True)
        self.compiler = compiler.Compiler(analyzer_options=analyzer_options,
                                          xspt_mode=False,
                                          compiler_stack_traces=True)

    def assign_after_filter_fails(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #set $foo = 'foo'
          $foo
          #set $foo = 'bar'
          $foo
        #end def
        """

        ast_root, function_node = self._build_function_template()
        first_assign = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode('foo'))
        function_node.append(first_assign)
        first_use = ast.FilterNode(ast.IdentifierNode('foo'))
        function_node.append(first_use)
        second_assign = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode('bar'))
        function_node.append(second_assign)
        second_use = ast.FilterNode(ast.IdentifierNode('foo'))
        function_node.append(second_use)

        optimization_analyzer = optimizer.OptimizationAnalyzer(
            ast_root, self.compiler.analyzer_options, self.compiler)

        optimization_analyzer.visit_ast = test_util.RecordedFunction(
            optimization_analyzer.visit_ast)

        self.assertRaises(compiler.Warning, optimization_analyzer.visit_ast,
                          ast_root)

    def double_assign_ok(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #set $foo = 'foo'
          #set $foo = 'bar'
          $foo
        #end def
        """

        ast_root, function_node = self._build_function_template()
        first_assign = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode('foo'))
        function_node.append(first_assign)
        second_assign = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode('bar'))
        function_node.append(second_assign)
        first_use = ast.FilterNode(ast.IdentifierNode('foo'))
        function_node.append(first_use)

        optimization_analyzer = optimizer.OptimizationAnalyzer(
            ast_root, self.compiler.analyzer_options, self.compiler)

        optimization_analyzer.visit_ast = test_util.RecordedFunction(
            optimization_analyzer.visit_ast)

        try:
            optimization_analyzer.visit_ast(ast_root)
        except compiler.Warning:
            self.fail('visit_ast raised WarningError unexpectedly.')


class TestPartialLocalIdentifiers(BaseTest):

    def setUp(self):
        # TODO: Use BaseTest.setUp()?
        analyzer_options = options.default_options
        analyzer_options.update(static_analysis=True,
                                directly_access_defined_variables=True)
        self.compiler = compiler.Compiler(analyzer_options=analyzer_options,
                                          xspt_mode=False,
                                          compiler_stack_traces=True)

    def test_simple_if(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = 1
          #end if
          $foo
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        assign_node = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1))
        if_node.append(assign_node)
        function_node.append(ast.PlaceholderNode('foo'))

        optimization_analyzer = self._get_analyzer(ast_root)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          optimization_analyzer.visit_ast, ast_root)

    def test_if_partial_else(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = 1
          #else
            #set $bar = 1
          #end if
          $foo
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        if_node.else_.append(ast.AssignNode(
            ast.IdentifierNode('bar'), ast.LiteralNode(1)))
        function_node.append(ast.PlaceholderNode('foo'))

        optimization_analyzer = self._get_analyzer(ast_root)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          optimization_analyzer.visit_ast, ast_root)

    def test_partial_if_else(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = 1
          #else
            #set $bar = 1
          #end if
          $bar
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        if_node.else_.append(ast.AssignNode(
            ast.IdentifierNode('bar'), ast.LiteralNode(1)))
        function_node.append(ast.PlaceholderNode('bar'))

        optimization_analyzer = self._get_analyzer(ast_root)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          optimization_analyzer.visit_ast, ast_root)

    def test_nested_else(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = 1
          #elif
            #set $foo = 2
          #else
            #set $foo = 3
          #end if
          $foo
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        if_node_2 = ast.IfNode(ast.LiteralNode(True))
        if_node_2.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(2)))
        if_node_2.else_.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(3)))
        if_node.else_.append(if_node_2)

        function_node.append(ast.PlaceholderNode('foo'))

        optimization_analyzer = self._get_analyzer(ast_root)

        try:
            optimization_analyzer.visit_ast(ast_root)
        except analyzer.SemanticAnalyzerError:
            self.fail('visit_ast raised SemanticAnalyzerError unexpectedly.')

    def test_nested_if(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #if True
              #set $foo = 1
            #else
              #set $foo = 2
            #end if
          #else
            #set $foo = 3
          #end if
          $foo
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        if_node_2 = ast.IfNode(ast.LiteralNode(True))
        if_node_2.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        if_node_2.else_.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(2)))
        if_node.append(if_node_2)
        if_node.else_.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(3)))
        function_node.append(ast.PlaceholderNode('foo'))

        optimization_analyzer = self._get_analyzer(ast_root)

        try:
            optimization_analyzer.visit_ast(ast_root)
        except analyzer.SemanticAnalyzerError:
            self.fail('visit_ast raised SemanticAnalyzerError unexpectedly.')

    def test_partial_nested_if(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #if True
              #set $foo = 1
            #else
              #set $bar = 2
            #end if
          #else
            #set $foo = 3
          #end if
          $foo
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        if_node_2 = ast.IfNode(ast.LiteralNode(True))
        if_node_2.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        if_node_2.else_.append(ast.AssignNode(
            ast.IdentifierNode('bar'), ast.LiteralNode(2)))
        if_node.append(if_node_2)
        if_node.else_.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(3)))
        function_node.append(ast.PlaceholderNode('foo'))

        optimization_analyzer = self._get_analyzer(ast_root)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          optimization_analyzer.visit_ast, ast_root)

    def test_partial_nested_else(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = 1
          #else
            #if
              #set $bar = 2
            #else
              #set $baz = 3
            #end if
          #end if
          $baz
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        if_node_2 = ast.IfNode(ast.LiteralNode(True))
        if_node_2.append(ast.AssignNode(
            ast.IdentifierNode('bar'), ast.LiteralNode(2)))
        if_node_2.else_.append(ast.AssignNode(
            ast.IdentifierNode('baz'), ast.LiteralNode(3)))
        if_node.else_.append(if_node_2)
        function_node.append(ast.PlaceholderNode('baz'))

        optimization_analyzer = self._get_analyzer(ast_root)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          optimization_analyzer.visit_ast, ast_root)

    def test_partial_nested_else_if(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = 1
          #else
            #if True
              #set $foo = 2
            #end if
          #end if
          $foo
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        if_node_2 = ast.IfNode(ast.LiteralNode(True))
        if_node_2.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(2)))
        if_node.else_.append(if_node_2)
        function_node.append(ast.PlaceholderNode('foo'))

        optimization_analyzer = self._get_analyzer(ast_root)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          optimization_analyzer.visit_ast, ast_root)

    def test_nested_else(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = 1
          #else
            #if
              #set $foo = 2
            #else
              #set $foo = 3
            #end if
          #end if
          $foo
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        if_node_2 = ast.IfNode(ast.LiteralNode(True))
        if_node_2.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(2)))
        if_node_2.else_.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(3)))
        if_node.else_.append(if_node_2)
        function_node.append(ast.PlaceholderNode('foo'))

        optimization_analyzer = self._get_analyzer(ast_root)

        try:
            optimization_analyzer.visit_ast(ast_root)
        except analyzer.SemanticAnalyzerError:
            self.fail('visit_ast raised SemanticAnalyzerError unexpectedly.')

    def test_nested_partial_use(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #if True
            #set $foo = 1
          #end if
          #if True
            $foo
          #end if
        #end def
        """

        ast_root, function_node, if_node = self._build_if_template()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        if_node_2 = ast.IfNode(ast.LiteralNode(True))
        if_node_2.append(ast.PlaceholderNode('foo'))
        function_node.append(if_node_2)

        optimization_analyzer = self._get_analyzer(ast_root)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          optimization_analyzer.visit_ast, ast_root)


class TestFinalPassHoistConditional(BaseTest):

    def setUp(self):
        analyzer_options = options.default_options
        analyzer_options.update(static_analysis=True,
                                directly_access_defined_variables=True,
                                hoist_conditional_aliases=True,
                                cache_filtered_placeholders=True)
        self.compiler = compiler.Compiler(analyzer_options=analyzer_options,
                                          xspt_mode=False,
                                          compiler_stack_traces=True)

    def test_hoist_both(self):
        self.ast_description = """
        file: TestTemplate
        #global $foo
        #def test_function
          #if True
            $foo
          #else
            $foo
          #end if
        #end def
        """

        def scope_setter(scope):
            scope.local_identifiers.add(ast.IdentifierNode('_rph_foo'))
            scope.aliased_expression_map[ast.PlaceholderNode('foo')] = (
                ast.IdentifierNode('_rph_foo'))
            scope.aliased_expression_map[ast.FilterNode(ast.IdentifierNode(
                '_rph_foo'))] = (ast.IdentifierNode('_fph123'))
            scope.alias_name_set.add('_fph123')
            scope.alias_name_set.add('_rph_foo')

        def build_conditional_body(node):
            node.append(ast.AssignNode(
                ast.IdentifierNode('_rph_foo'), ast.PlaceholderNode('foo')))
            node.append(ast.AssignNode(
                ast.IdentifierNode('_fph123'), ast.FilterNode(
                    ast.IdentifierNode('_rph_foo'))))
            node.append(ast.BufferWrite(ast.IdentifierNode('_fph123')))

        ast_root, function_node, if_node = self._build_if_template()
        ast_root.global_placeholders.add('foo')
        scope_setter(function_node.scope)
        function_node.scope.local_identifiers.add(ast.IdentifierNode('self'))
        scope_setter(if_node.scope)
        scope_setter(if_node.else_.scope)
        build_conditional_body(if_node)
        build_conditional_body(if_node.else_)

        final_pass_analyzer = optimizer.FinalPassAnalyzer(
            ast_root, self.compiler.analyzer_options, self.compiler)

        final_pass_analyzer.hoist = test_util.RecordedFunction(
            final_pass_analyzer.hoist)

        final_pass_analyzer.visit_ast(ast_root)

        # The 4 calls are hoisting the rph alias and the fph alias out of
        # both the if and else clauses.
        self.assertEqual(len(final_pass_analyzer.hoist.GetCalls()), 4)


class TestHoistPlaceholders(BaseTest):

    def setUp(self):
        analyzer_options = options.default_options
        analyzer_options.update(cache_resolved_placeholders=True,
                                enable_warnings=True,
                                warnings_as_errors=True,
                                directly_access_defined_variables=True,
                                static_analysis=False)
        self.compiler = compiler.Compiler(analyzer_options=analyzer_options,
                                          xspt_mode=False,
                                          compiler_stack_traces=True)

    def fake_placeholdernode_replacement(self, placeholder, local_var,
                                         cached_placeholder, local_identifiers):
        return self.analyzer_options.cache_resolved_placeholders

    def _get_analyzer_and_visit(self, ast_root):
        analyzer = self._get_analyzer(ast_root)
        analyzer._placeholdernode_replacement = test_util.RecordedFunction(
            self.fake_placeholdernode_replacement)
        analyzer.visit_ast(ast_root)
        return analyzer

    def test_simple_hoist(self):
        self.ast_description = """
        file: TestTemplate

        #def test_function
          $foo
          $foo
        #end def
        """

        ast_root, function_node = self._build_function_template()
        function_node.append(ast.PlaceholderNode('foo'))
        function_node.append(ast.PlaceholderNode('foo'))

        optimization_analyzer = self._get_analyzer_and_visit(ast_root)
        self.assertEqual(
            optimization_analyzer._placeholdernode_replacement.GetResults(),
            [True, True])

    def test_hoists_both_from_plus(self):
        self.ast_description = """
        file: TestTemplate

        #global $foo

        #def test_function
          #set $bar = $foo + $foo
        #end def
        """

        ast_root, function_node = self._build_function_template()
        ast_root.global_placeholders.add('foo')
        function_node.append(ast.AssignNode(
            ast.IdentifierNode('bar'), ast.BinOpNode('+', ast.PlaceholderNode(
                'foo'), ast.PlaceholderNode('foo'))))

        optimization_analyzer = self._get_analyzer_and_visit(ast_root)
        self.assertEqual(
            optimization_analyzer._placeholdernode_replacement.GetResults(),
            [True, True])

    def test_hoists_lhs_only_from_and(self):
        self.ast_description = """
        file: TestTemplate

        #def test_function
          #if $foo or $bar
          #end if
        #end def
        """

        condition = ast.BinOpNode('or', ast.PlaceholderNode('foo'),
                                  ast.PlaceholderNode('bar'))
        ast_root, function_node, if_node = self._build_if_template(condition)

        optimization_analyzer = self._get_analyzer_and_visit(ast_root)
        self.assertEqual(
            optimization_analyzer._placeholdernode_replacement.GetResults(),
            [True, False])


class TestAssignSlice(BaseTest):

    def test_index_before_assign_error(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #set $foo[1] = 1
        #end def
        """

        ast_root, function_node = self._build_function_template()
        assign_node = ast.AssignNode(
            ast.SliceNode(
                ast.IdentifierNode('foo'), ast.LiteralNode(1)),
            ast.LiteralNode(1))
        function_node.append(assign_node)

        optimization_analyzer = self._get_analyzer(ast_root)
        self.assertRaises(analyzer.SemanticAnalyzerError,
                          optimization_analyzer.visit_ast, ast_root)

    def test_index_after_assign_ok(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #set $foo = {}
          #set $foo[1] = 1
        #end def
        """

        ast_root, function_node = self._build_function_template()
        assign_node1 = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.DictLiteralNode())
        function_node.append(assign_node1)
        assign_node2 = ast.AssignNode(
            ast.SliceNode(
                ast.IdentifierNode('foo'), ast.LiteralNode(1)),
            ast.LiteralNode(1))
        function_node.append(assign_node2)

        optimization_analyzer = self._get_analyzer(ast_root)

        try:
            optimization_analyzer.visit_ast(ast_root)
        except analyzer.SemanticAnalyzerError:
            self.fail('visit_ast raised SemanticAnalyzerError unexpectedly.')

    def test_index_scope_ok(self):
        self.ast_description = """
        file: TestTemplate
        #def test_function
          #set $foo = {}
          #if True
            #set $foo[1] = 1
          #end if
        #end def
        """

        ast_root, function_node = self._build_function_template()
        assign_node1 = ast.AssignNode(
            ast.IdentifierNode('foo'), ast.DictLiteralNode())
        function_node.append(assign_node1)
        if_node = ast.IfNode(ast.LiteralNode(True))
        assign_node2 = ast.AssignNode(
            ast.SliceNode(
                ast.IdentifierNode('foo'), ast.LiteralNode(1)),
            ast.LiteralNode(1))
        if_node.append(assign_node2)
        function_node.append(if_node)

        optimization_analyzer = self._get_analyzer(ast_root)

        try:
            optimization_analyzer.visit_ast(ast_root)
        except analyzer.SemanticAnalyzerError:
            self.fail('visit_ast raised SemanticAnalyzerError unexpectedly.')


class TestCollectWrites(BaseTest):

    def setUp(self):
        analyzer_options = options.default_options
        analyzer_options.update(cache_resolved_placeholders=True,
                                enable_warnings=True,
                                warnings_as_errors=True,
                                directly_access_defined_variables=True,
                                static_analysis=False,
                                batch_buffer_writes=True)
        self.compiler = compiler.Compiler(analyzer_options=analyzer_options,
                                          xspt_mode=False,
                                          compiler_stack_traces=True)

    def _get_analyzer(self, ast_root):
        optimization_analyzer = optimizer.FinalPassAnalyzer(
            ast_root, self.compiler.analyzer_options, self.compiler)
        optimization_analyzer.visit_ast = test_util.RecordedFunction(
            optimization_analyzer.visit_ast)
        return optimization_analyzer

    def test_collect_writes_no_change(self):
        self.ast_description = """
        file: TestTemplate

        #def test_function
          foo
        #end def
        """

        ast_root, function_node = self._build_function_template()
        function_node.append(ast.BufferWrite(ast.LiteralNode('foo')))
        expected_hash = hash(ast_root)

        optimization_analyzer = self._get_analyzer(ast_root)
        got_hash = hash(optimization_analyzer.optimize_ast())
        self.assertEqual(expected_hash, got_hash)

    def test_collect_writes_join_simple(self):
        self.ast_description = """
        file: TestTemplate

        #def test_function
          foo
          bar
        #end def
        """

        ast_root, function_node = self._build_function_template()
        function_node.append(ast.BufferWrite(ast.LiteralNode('foo')))
        function_node.append(ast.BufferWrite(ast.LiteralNode('bar')))
        optimization_analyzer = self._get_analyzer(ast_root)

        ast_root, function_node = self._build_function_template()
        tuple_node = ast.TupleLiteralNode()
        tuple_node.append(ast.LiteralNode('foo'))
        tuple_node.append(ast.LiteralNode('bar'))
        function_node.append(ast.BufferExtend(tuple_node))
        expected_hash = hash(ast_root)

        got_hash = hash(optimization_analyzer.optimize_ast())
        self.assertEqual(expected_hash, got_hash)

    def test_collect_writes_join_if(self):
        self.ast_description = """
        file: TestTemplate

        #def test_function
          foo
          bar
          #if True
            #set $foo = 1
          #end if
          baz
          boo
        #end def
        """

        ast_root, function_node = self._build_function_template()
        function_node.append(ast.BufferWrite(ast.LiteralNode('foo')))
        function_node.append(ast.BufferWrite(ast.LiteralNode('bar')))
        if_node = ast.IfNode()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        function_node.append(if_node)
        function_node.append(ast.BufferWrite(ast.LiteralNode('baz')))
        function_node.append(ast.BufferWrite(ast.LiteralNode('boo')))
        optimization_analyzer = self._get_analyzer(ast_root)

        ast_root, function_node = self._build_function_template()
        tuple_node = ast.TupleLiteralNode()
        tuple_node.append(ast.LiteralNode('foo'))
        tuple_node.append(ast.LiteralNode('bar'))
        function_node.append(ast.BufferExtend(tuple_node))
        if_node = ast.IfNode()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        function_node.append(if_node)
        tuple_node = ast.TupleLiteralNode()
        tuple_node.append(ast.LiteralNode('baz'))
        tuple_node.append(ast.LiteralNode('boo'))
        function_node.append(ast.BufferExtend(tuple_node))

        expected_hash = hash(ast_root)

        got_hash = hash(optimization_analyzer.optimize_ast())
        self.assertEqual(expected_hash, got_hash)

    def test_duplicate_node_collect(self):
        self.ast_description = """
        file: TestTemplate

        #def test_function
          foo
          bar
          #if True
            #set $foo = 1
          #end if
          baz
          boo
          #if True
            #set $foo = 1
          #end if
        #end def

        NOTE: This test will break if collect_writes is written
        using ast.ASTNode.insert_before.
        """

        ast_root, function_node = self._build_function_template()
        function_node.append(ast.BufferWrite(ast.LiteralNode('foo')))
        function_node.append(ast.BufferWrite(ast.LiteralNode('bar')))
        if_node = ast.IfNode()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        function_node.append(if_node)
        function_node.append(ast.BufferWrite(ast.LiteralNode('baz')))
        function_node.append(ast.BufferWrite(ast.LiteralNode('boo')))
        if_node = ast.IfNode()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        optimization_analyzer = self._get_analyzer(ast_root)

        ast_root, function_node = self._build_function_template()
        tuple_node = ast.TupleLiteralNode()
        tuple_node.append(ast.LiteralNode('foo'))
        tuple_node.append(ast.LiteralNode('bar'))
        function_node.append(ast.BufferExtend(tuple_node))
        if_node = ast.IfNode()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))
        function_node.append(if_node)
        tuple_node = ast.TupleLiteralNode()
        tuple_node.append(ast.LiteralNode('baz'))
        tuple_node.append(ast.LiteralNode('boo'))
        function_node.append(ast.BufferExtend(tuple_node))
        if_node = ast.IfNode()
        if_node.append(ast.AssignNode(
            ast.IdentifierNode('foo'), ast.LiteralNode(1)))

        expected_hash = hash(ast_root)

        got_hash = hash(optimization_analyzer.optimize_ast())
        self.assertEqual(expected_hash, got_hash)


class TestDoNode(BaseTest):

    def test_do_placeholder_replace(self):
        code = """
        #global $bar

        #def foo
          #do $bar
        #end def
        """

        ast_root = self._compile(code)
        semantic_analyzer = analyzer.SemanticAnalyzer(
            'TestTemplate', ast_root, self.compiler.analyzer_options,
            self.compiler)
        analyzed_tree = semantic_analyzer.get_ast()

        optimization_analyzer = self._get_analyzer(analyzed_tree)
        optimization_analyzer.optimize_ast()


class TestCacheFilterArgs(BaseTest):

    def setUp(self):
        analyzer_options = options.default_options
        analyzer_options.update(cache_resolved_udn_expressions=True,
                                enable_warnings=True,
                                warnings_as_errors=True,
                                directly_access_defined_variables=True,
                                static_analysis=False)
        self.compiler = compiler.Compiler(analyzer_options=analyzer_options,
                                          xspt_mode=False,
                                          compiler_stack_traces=True)

    def test_cache_filter_args_udn(self):
        code = """
        #from foo import bar

        #def func
          $bar.baz('arg')
        #end def
        """

        ast_root = self._compile(code)
        semantic_analyzer = analyzer.SemanticAnalyzer(
            'TestTemplate', ast_root, self.compiler.analyzer_options,
            self.compiler)
        analyzed_tree = semantic_analyzer.get_ast()

        optimization_analyzer = self._get_analyzer(analyzed_tree)
        optimized_tree = optimization_analyzer.optimize_ast()

        def pred(node):
            return type(node) == ast.AssignNode

        alias_assign = walker.find_node(optimized_tree, pred)
        if not alias_assign:
            self.fail('There should be an ast.AssignNode due to caching')

    def test_cache_filter_args_identifier(self):
        code = """
        #implements library
        #from foo import library bar

        #def func
          $bar.baz('arg')
        #end def
        """

        ast_root = self._compile(code)
        semantic_analyzer = analyzer.SemanticAnalyzer(
            'TestTemplate', ast_root, self.compiler.analyzer_options,
            self.compiler)
        analyzed_tree = semantic_analyzer.get_ast()

        optimization_analyzer = self._get_analyzer(analyzed_tree)
        optimized_tree = optimization_analyzer.optimize_ast()

        def pred(node):
            return type(node) == ast.AssignNode
            alias_assign = walker.find_node(optimized_tree, pred)
            if not alias_assign:
                self.fail('There should be an ast.AssignNode due to caching')


class TestFilterInMacro(BaseTest):

    def test_filter_function_macro(self):
        code = """
        #def foo
          $my_macro()
        #end def
        """

        def macro_function(macro_node, arg_map, compiler):
            return '#set $bar = self.filter_function("test")\n'

        self.compiler.register_macro('macro_function_my_macro',
                                     macro_function,
                                     parse_rule='fragment_goal')

        ast_root = self._compile(code)
        semantic_analyzer = analyzer.SemanticAnalyzer(
            'TestTemplate', ast_root, self.compiler.analyzer_options,
            self.compiler)
        analyzed_tree = semantic_analyzer.get_ast()

        optimization_analyzer = self._get_analyzer(analyzed_tree)
        optimized_tree = optimization_analyzer.optimize_ast()

        def pred(node):
            return bool(type(node) == ast.IdentifierNode and node.name ==
                        '_self_filter_function')

        filter_node = walker.find_node(optimized_tree, pred)
        if not filter_node:
            self.fail('Expected _self_filter_function in ast')

    def test_private_filter_function_macro(self):
        code = """
        #def foo
          $my_macro()
        #end def
        """

        def macro_function(macro_node, arg_map, compiler):
            return '#set $bar = self._filter_function("test")\n'

        self.compiler.register_macro('macro_function_my_macro',
                                     macro_function,
                                     parse_rule='fragment_goal')

        ast_root = self._compile(code)
        semantic_analyzer = analyzer.SemanticAnalyzer(
            'TestTemplate', ast_root, self.compiler.analyzer_options,
            self.compiler)
        analyzed_tree = semantic_analyzer.get_ast()

        optimization_analyzer = self._get_analyzer(analyzed_tree)
        optimized_tree = optimization_analyzer.optimize_ast()

        def pred(node):
            return bool(type(node) == ast.IdentifierNode and node.name ==
                        '_self_private_filter_function')

        filter_node = walker.find_node(optimized_tree, pred)
        if not filter_node:
            self.fail('Expected _self_private_filter_function in ast')


class TestHoistOnlyClean(BaseTest):

    def setUp(self):
        analyzer_options = options.o3_options
        self.compiler = compiler.Compiler(analyzer_options=analyzer_options,
                                          xspt_mode=False,
                                          compiler_stack_traces=True)
        self.compiler.new_registry_format = True
        self.compiler.function_name_registry['reg_f'] = ('a.reg_f',
                                                         ['skip_filter'])

    def _get_final_tree(self, code):
        ast_root = self._compile(code)
        semantic_analyzer = analyzer.SemanticAnalyzer(
            'TestTemplate', ast_root, self.compiler.analyzer_options,
            self.compiler)
        analyzed_tree = semantic_analyzer.get_ast()

        optimization_analyzer = self._get_analyzer(analyzed_tree)
        optimized_tree = optimization_analyzer.optimize_ast()

        final_pass_analyzer = optimizer.FinalPassAnalyzer(
            optimized_tree, self.compiler.analyzer_options, self.compiler)

        return final_pass_analyzer.optimize_ast()

    def test_should_hoist_for(self):
        code = """
        #def foo($bar)
          #for $i in []
            $reg_f($bar)
          #end for
        #end def
        """

        final_tree = self._get_final_tree(code)

        def pred(node):
            return type(node) == ast.AssignNode and type(
                node.parent) == ast.FunctionNode

        alias = walker.find_node(final_tree, pred)
        if not alias:
            self.fail(
                'Expected to find ast.AssignNode hoisted to function scope.')

    def test_should_not_hoist_for(self):
        code = """
        #def foo($bar)
          #for $i in []
            #set $bar["1"] = 1
            $reg_f($bar)
          #end for
        #end def
        """

        final_tree = self._get_final_tree(code)

        def pred(node):
            return type(node) == ast.AssignNode and type(
                node.parent) == ast.FunctionNode

        alias = walker.find_node(final_tree, pred)
        if alias:
            self.fail('ast.AssignNode should not be hoisted to function scope.')

    def test_should_hoist_if(self):
        code = """
        #def foo($bar)
          #if True
            $reg_f($bar)
          #else
            $reg_f($bar)
          #end if
        #end def
          """

        final_tree = self._get_final_tree(code)

        def pred(node):
            return type(node) == ast.AssignNode and type(
                node.parent) == ast.FunctionNode

        alias = walker.find_node(final_tree, pred)
        if not alias:
            self.fail(
                'Expected to find ast.AssignNode hoisted to function scope.')

    def test_should_not_hoist_if_do(self):
        code = """
        #def f
        #end def
        #def foo($bar)
          #if True
            #do $f($bar)
            $reg_f($bar)
          #else
            $reg_f($bar)
          #end if
        #end def
        """

        final_tree = self._get_final_tree(code)

        def pred_if(node):
            return type(node) == ast.IfNode

        if_node = walker.find_node(final_tree, pred_if)

        def pred(node):
            return type(node) == ast.AssignNode

        alias = walker.find_node(if_node, pred)
        if not alias:
            self.fail('ast.AssignNode should be present in the If block.')

    def test_should_not_hoist_if_set(self):
        code = """
        #def foo($bar)
          #if True
            #set $bar[1] = 1
            $reg_f($bar)
          #else
            $reg_f($bar)
          #end if
        #end def
        """

        final_tree = self._get_final_tree(code)

        def pred(node):
            return type(node) == ast.AssignNode and type(
                node.parent) == ast.FunctionNode

        alias = walker.find_node(final_tree, pred)
        if alias:
            self.fail(
                'ast.AssignNode should not be hoisted to the ast.FunctionNode.')

    def test_should_not_hoist_if_set_output(self):
        code = """
        #def foo($bar)
          #if True
            #set $bar[1] = 1
            $bar
          #else
            $bar
          #end if
        #end def
        """

        final_tree = self._get_final_tree(code)

        def pred(node):
            return type(node) == ast.AssignNode and type(
                node.parent) == ast.FunctionNode

        alias = walker.find_node(final_tree, pred)
        if alias:
            self.fail(
                'ast.AssignNode should not be hoisted to the ast.FunctionNode.')


class TestSanitizationOptimizations(BaseTest):

    def setUp(self):
        analyzer_options = options.default_options
        self.compiler = compiler.Compiler(analyzer_options=analyzer_options,
                                          xspt_mode=False,
                                          compiler_stack_traces=True,
                                          baked_mode=True)

    def _get_optimized_tree(self, code):
        ast_root = self._compile(code)
        semantic_analyzer = analyzer.SemanticAnalyzer(
            'TestTemplate', ast_root, self.compiler.analyzer_options,
            self.compiler)
        analyzed_tree = semantic_analyzer.get_ast()

        optimization_analyzer = self._get_analyzer(analyzed_tree)
        return optimization_analyzer.optimize_ast()

    def test_should_not_need_sanitization_if(self):
        code = """
        #def foo($bar)
          #if $bar.baz()
            BLAH
          #end if
        #end def
        """

        optimized_tree = self._get_optimized_tree(code)

        def pred(node):
            return type(node) == ast.CallFunctionNode

        call_node = walker.find_node(optimized_tree, pred)
        if not call_node:
            self.fail('Expected to find a ast.CallFunctionNode.')
        if call_node.sanitization_state != ast.SanitizedState.NOT_OUTPUTTED:
            self.fail(
                'Expected node in test expression to not need sanitization.')

    def test_should_not_need_sanitization_do(self):
        code = """
        #def foo($bar)
          #do $bar.baz()
        #end def
        """

        optimized_tree = self._get_optimized_tree(code)

        def pred(node):
            return type(node) == ast.CallFunctionNode

        call_node = walker.find_node(optimized_tree, pred)
        if not call_node:
            self.fail('Expected to find a ast.CallFunctionNode.')
        if call_node.sanitization_state != ast.SanitizedState.NOT_OUTPUTTED:
            self.fail('Expected node in #do to not need sanitization.')

    def test_should_not_need_sanitization_filter(self):
        code = """
        #def foo($bar)
          $bar()
        #end def
        """

        optimized_tree = self._get_optimized_tree(code)

        def pred(node):
            return (type(node) == ast.CallFunctionNode and
                    type(node.parent) == ast.FilterNode)

        call_node = walker.find_node(optimized_tree, pred)
        if not call_node:
            self.fail('Expected to find a ast.CallFunctionNode.')
        if (call_node.sanitization_state !=
                ast.SanitizedState.OUTPUTTED_IMMEDIATELY):
            self.fail(
                'Expected node in ast.FilterNode to not need sanitization.')


if __name__ == '__main__':
    unittest.main()



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.


class RepeatTracker(object):

    def __init__(self):
        self.repeater_map = {}

    def __setitem__(self, key, value):
        try:
            self.repeater_map[key].index = value
        except KeyError as e:
            self.repeater_map[key] = Repeater(value)

    def __getitem__(self, key):
        return self.repeater_map[key]


class Repeater(object):

    def __init__(self, index=0, item=None, length=None):
        self.index = index
        self.item = item
        self.length = length

    @property
    def number(self):
        return self.index + 1

    @property
    def even(self):
        return not (self.index % 2)

    @property
    def odd(self):
        return (self.index % 2)

    @property
    def first(self):
        return (self.index == 0)

    @property
    def last(self):
        return (self.index == (self.length - 1))


class RepeatIterator(object):

    def __init__(self, iterable):
        self.src_iterable = iterable
        self.src_iterator = enumerate(iterable)
        try:
            self.length = len(iterable)
        except TypeError:
            # if the iterable is a generator, then we have no length
            self.length = None

    def __iter__(self):
        return self

    def __next__(self):
        index, item = next(self.src_iterator)
        return Repeater(index, item, self.length)



'''
Spitfire port for Python 3
'''

# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

__author__ = 'Nafi Amaan Hossain' # Originally by Mike Solomon
__author_email__ = '<nafines007@gmail.com>'
__version__ = '0.00.1.0.1'
__license__ = 'BSD License'

import spitfire
import spitfire.compiler.util
import spitfire.compiler.options
import os

class Environment():
	def __init__(self, dirpath: str, config={}):
		self.home = dirpath

	def render(self, filename: str, opts: [{}], template_name="_", enable_filters=False):

		o3_opts = spitfire.compiler.options.o3_options

		file = open(self.home + '/' + filename)

		tmpl_o3 = spitfire.compiler.util.load_template(file.read(),
		                                               'ample',
		                                               analyzer_options=o3_opts)
		return tmpl_o3(search_list=[{'table': opts}]).main()







# Copyright 2016 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from builtins import object
import sys


class RecordedExcInfo(object):
    """Container for exception info recorded by the RecordedFunction class."""

    def __init__(self, exc_info):
        """Initializer.

        Args:
        exc_info: Tuple returned by sys.exc_info()
        """
        self._exc_info = exc_info

    def GetExcInfo(self):
        """Returns the tuple returned by sys.exc_info() stored in this
        object."""
        return self._exc_info

    def GetException(self):
        """Returns the exception object thrown by a wrapped function."""
        exc_type, exc_obj, unused_traceback = self._exc_info

        # Correctly handles exceptions that are not Exception sub-classes.
        if exc_obj is not None:
            return exc_obj
        return exc_type


class RecordedFunction(object):
    """Records the parameters and results of every call to a function.

    Use it like this:

      def power(num, exp=2):
        return num ** exp

      recorded_power = RecordedFunction(power)
      self.assertEqual(recorded_power(2), 4)
      self.assertEqual(recorded_power(2, 3), 8)
      self.assertEqual(recorded_power(5, exp=1), 5)

      # Look at how it was called and how it responded.
      expected_calls = [
        ((2,), {}), # first call
        ((2, 3), {}), # second call
        ((5,), {"exp": 1}) # third call
      ]
      self.assertListEqual(recorded_power.GetCalls(), expected_calls)

      expected_results = [
        4,
        8,
        5
      ]
      self.assertListEqual(recorded_power.GetResults(), expected_results)

    If at any time an exception is thrown by the wrapped function, an instance
    of the RecordedExcInfo class that contains the exception will be placed in
    the results list. This ensures that the indexes of GetCalls() and
    GetResults() match up regardless of what happens.

    Example:

      def thrower(e):
        raise e

      recorded_thrower = RecordedFunction(thrower)

      class MyError(Exception): pass

      expected_exception = MyError()

      try:
        recorded_thrower(expected_exception)
      except MyError as e:
        self.assertIs(e, expected_exception)

      # Look at how it was called and how it responded.
      expected_calls = [
        ((expected_exception,), {})
      ]
      self.assertListEqual(recorded_thrower.GetCalls(), expected_calls)

      results = recorded_thrower.GetResults()

      self.assertEqual(len(results), 1)
      self.assertIsInstance(results[0], RecordedExcInfo)
      self.assertIs(results[0].GetException(), expected_exception)
    """

    def __init__(self, func):
        """Initializer.

        Args:
          func: Callable function to wrap with the recorder.
        """
        self._calls = []
        self._results = []
        self._func = func

    def __call__(self, *args, **kwargs):
        """Call the function with the supplied parameter, recording results.

        Args:
          *args: Positional arguments.
          **kwargs: Keyword arguments.

        Returns:
          Result of calling the wrapped function with the supplied parameters.
        """
        self._calls.append((args, kwargs))

        try:
            result = self._func(*args, **kwargs)
            self._results.append(result)
            return result
        except:
            self._results.append(RecordedExcInfo(sys.exc_info()))
            raise

    def GetCalls(self):
        """Get the list of calls made so far.

        Returns:
          Each entry in the list will be a tuple (args, kwargs) where:
            args: Tuple of positional arguments.
            kwargs: Dictionary of keyword arguments.
        """
        # Use a copy so callers can't modify the original.
        return self._calls[:]

    def GetResults(self):
        """Collect the list of results returned by the wrapped function.

        When the wrapped function raises an exception, an instance of the
        RecordedExcInfo class containing the exception will be placed in the
        results list since no actual value is available.

        Returns:
          The list of results returned by the wrapped function.
        """
        # Use a copy so callers can't modify the original.
        return self._results[:]



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

from builtins import zip
import re
import string
import unicodedata

normal_characters = string.ascii_lowercase + string.ascii_uppercase
mangled_character_names = [
    'LATIN SMALL LETTER A WITH RING ABOVE',
    'LATIN SMALL LETTER THORN',
    'LATIN SMALL LETTER C WITH CIRCUMFLEX',
    'LATIN SMALL LETTER D WITH CARON',
    'LATIN SMALL LETTER E WITH GRAVE',
    'LATIN SMALL LETTER F WITH HOOK',
    'LATIN SMALL LETTER G WITH CIRCUMFLEX',
    'LATIN SMALL LETTER H WITH STROKE',
    'LATIN SMALL LETTER I WITH TILDE',
    'LATIN SMALL LETTER J WITH CIRCUMFLEX',
    'LATIN SMALL LETTER K WITH CEDILLA',
    'LATIN SMALL LETTER L WITH ACUTE',
    'SCRIPT CAPITAL M',
    'LATIN SMALL LETTER N WITH ACUTE',
    'LATIN SMALL LETTER ETH',
    'GREEK SMALL LETTER RHO',
    'LATIN SMALL LETTER O WITH OGONEK',
    'LATIN SMALL LETTER R WITH CARON',
    'LATIN SMALL LETTER S WITH ACUTE',
    'LATIN SMALL LETTER T WITH STROKE',
    'LATIN SMALL LETTER U WITH DOUBLE GRAVE',
    'LATIN SMALL LETTER V WITH TILDE',
    'LATIN SMALL LETTER W WITH DOT BELOW',
    'LATIN SMALL LETTER X WITH DOT ABOVE',
    'GREEK SMALL LETTER GAMMA',
    'LATIN SMALL LETTER Z WITH HOOK',
    'LATIN CAPITAL LETTER A WITH CIRCUMFLEX',
    'LATIN SMALL LETTER SHARP S',
    'LATIN CAPITAL LETTER C WITH ACUTE',
    'LATIN CAPITAL LETTER D WITH STROKE',
    'LATIN CAPITAL LETTER E WITH MACRON',
    'LATIN CAPITAL LETTER F WITH DOT ABOVE',
    'LATIN CAPITAL LETTER G WITH CIRCUMFLEX',
    'LATIN CAPITAL LETTER H WITH CIRCUMFLEX',
    'LATIN CAPITAL LETTER I WITH MACRON',
    'LATIN CAPITAL LETTER J WITH CIRCUMFLEX',
    'LATIN CAPITAL LETTER K WITH CEDILLA',
    'LATIN CAPITAL LETTER L WITH ACUTE',
    'LATIN CAPITAL LETTER M WITH DOT ABOVE',
    'LATIN CAPITAL LETTER N WITH ACUTE',
    'LATIN CAPITAL LETTER O WITH BREVE',
    'LATIN CAPITAL LETTER P WITH DOT ABOVE',
    'LATIN CAPITAL LETTER O WITH OGONEK AND MACRON',
    'LATIN CAPITAL LETTER R WITH CARON',
    'LATIN CAPITAL LETTER S WITH ACUTE',
    'LATIN CAPITAL LETTER T WITH STROKE',
    'LATIN CAPITAL LETTER U WITH TILDE',
    'LATIN CAPITAL LETTER V WITH TILDE',
    'LATIN CAPITAL LETTER W WITH CIRCUMFLEX',
    'CYRILLIC CAPITAL LETTER ZHE',
    'LATIN CAPITAL LETTER Y WITH CIRCUMFLEX',
    'LATIN CAPITAL LETTER Z WITH ACUTE',
]

char_map = dict(list(zip(normal_characters, [unicodedata.lookup(name)
                                        for name in mangled_character_names])))


# return a string with interesting unicode characters to make sure
# an application is handling unicode correctly
def i18n_mangled_message(msg):
    return ''.join([char_map.get(c, c) for c in msg])


whitespace_regex = re.compile('\s+', re.UNICODE)


def normalize_whitespace(text):
    return whitespace_regex.sub(' ', text)



# Copyright 2014 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import logging
import unittest

from spitfire.compiler import ast
from spitfire.compiler import util
from spitfire.compiler import walker
from third_party.yapps2 import yappsrt


class BaseTest(unittest.TestCase):

    def _compile(self, template_content):
        template_node = util.parse_template(template_content)
        template_node.source_path = 'test_template.spt'
        return template_node


class TestEscapeHash(BaseTest):

    @staticmethod
    def _def_foo_pred(node):
        return type(node) == ast.DefNode and node.name == 'foo'

    def test_escape_simple(self):
        code = """
        #def foo
        \#test
        #end def
        """

        template = self._compile(code)

        def_node = walker.find_node(template, TestEscapeHash._def_foo_pred)
        text = ''.join([node.value
                        for node in def_node.child_nodes
                        if type(node) == ast.TextNode])
        self.assertEqual(text, '#test')

    def test_backslash_hash_escape(self):
        code = """
        #def foo
        \\\#test
        #end def
        """

        template = self._compile(code)

        def_node = walker.find_node(template, TestEscapeHash._def_foo_pred)
        text = ''.join([node.value
                        for node in def_node.child_nodes
                        if type(node) == ast.TextNode])
        self.assertEqual(text, '\\#test')

    def test_escape_needed(self):
        code = """
        #def foo
        \#if
        #end def
        """

        template = self._compile(code)

        def_node = walker.find_node(template, TestEscapeHash._def_foo_pred)
        text = ''.join([node.value
                        for node in def_node.child_nodes
                        if type(node) == ast.TextNode])
        self.assertEqual(text, '#if')


class TestDo(BaseTest):

    @staticmethod
    def _do_pred(node):
        return type(node) == ast.DoNode

    def test_do_syntax(self):
        code = """
        #def foo
        #do $bar()
        #end def
        """

        template = self._compile(code)
        do_node = walker.find_node(template, TestDo._do_pred)
        if not do_node:
            self.fail('Do node should be present in AST')

    def test_bad_do_syntax(self):
        code = """
        #def foo
        #do #do $bar()
        #end def
        """

        # Temporarily disable logging to avoid spurious exception output.
        logging.disable(logging.ERROR)
        self.assertRaises(yappsrt.FatalParseError, self._compile, code)
        # Re-enable logging.
        logging.disable(logging.NOTSET)

    def test_do_expression(self):
        code = """
        #def foo
        #do 1 + 2
        #end def
        """

        template = self._compile(code)

        do_node = walker.find_node(template, TestDo._do_pred)
        if not do_node:
            self.fail('Do node should be present in AST')


class TestAllowRaw(BaseTest):

    @staticmethod
    def _allow_raw_pred(node):
        return type(node) == ast.AllowRawNode

    def test_allow_raw(self):
        code = """
        #allow_raw

        #def foo
        #end def
        """

        template = self._compile(code)

        allow_raw_node = walker.find_node(template,
                                          TestAllowRaw._allow_raw_pred)
        if not allow_raw_node:
            self.fail('ast.AllowRawNode should be present in AST')


if __name__ == '__main__':
    unittest.main()



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# resolve unified-dotted-notation and placeholders
# this means search objects and dictionaries in the same way
# using attribute-like syntax from python
# syntactically, 'name' will always be a valid identifier - so you won't get
# name='my attribute' - it must be a legal python identifier

import builtins
import inspect
import logging
import weakref

from spitfire import runtime
# Import the accelerated C module if available.
try:
    from spitfire.runtime import _udn
except ImportError:
    _udn = None

# Avoid extra lookups on critical paths by aliasing imports directly.
UndefinedAttribute = runtime.UndefinedAttribute
UndefinedPlaceholder = runtime.UndefinedPlaceholder
UnresolvedPlaceholder = runtime.UnresolvedPlaceholder


# create a sentinel value for missing attributes
class __MissingAttr(object):
    pass


MissingAttr = __MissingAttr()


# sentinel class, in case you want to have a default that is None
class __Unspecified(object):
    pass


Unspecified = __Unspecified()


# Cheetah supports autocalling - Spitfire does not. this stand-in class will
# raise an exception if you do something like compare a function object.
class CallOnlyPlaceholder(object):

    def __init__(self, name, function):
        self.name = name
        self.function = function

    def __call__(self, *pargs, **kargs):
        return self.function(*pargs, **kargs)

    @property
    def template_method(self):
        return getattr(self.function, 'template_method')

    @property
    def skip_filter(self):
        return getattr(self.function, 'skip_filter')

    def __cmp__(self, unused_other):
        raise runtime.PlaceholderError(self.name,
                                       'function placeholder was not called')

    def __bool__(self):
        raise runtime.PlaceholderError(self.name,
                                       'function placeholder was not called')


# TODO - optimize performance
def _resolve_udn_prefer_attr(_object, name, raise_exception=False):
    try:
        return getattr(_object, name)
    except AttributeError:
        try:
            return _object[name]
        except (KeyError, TypeError):
            if raise_exception:
                raise runtime.UDNResolveError(name, dir(_object))
            else:
                return UndefinedAttribute(name, dir(_object))


def _resolve_udn_prefer_dict(_object, name, raise_exception=False):
    try:
        return _object[name]
    except (KeyError, TypeError):
        try:
            return getattr(_object, name)
        except AttributeError:
            if raise_exception:
                raise runtime.UDNResolveError(name, dir(_object))
            else:
                return UndefinedAttribute(name, dir(_object))


# this is always faster than catching an exception when that exception isn't
# truly exceptional,  but semi-expected
# using a sentinel should be quicker than calling hasattr then getattr
# this is true when the expected hit rate on an attribute is relatively
# reasonable - say 50% chance
def _resolve_udn_prefer_attr2(_object, name, raise_exception=False):
    val = getattr(_object, name, MissingAttr)
    if val is not MissingAttr:
        return val
    try:
        return _object[name]
    except (KeyError, TypeError):
        if raise_exception:
            raise runtime.UDNResolveError(name, dir(_object))
        else:
            return UndefinedAttribute(name, dir(_object))


# this version is slightly faster when there are a lot of misses on attributes
def _resolve_udn_prefer_attr3(_object, name, raise_exception=False):
    if hasattr(_object, name):
        return getattr(_object, name)
    try:
        return _object[name]
    except (KeyError, TypeError):
        if raise_exception:
            raise runtime.UDNResolveError(name, dir(_object))
        else:
            return UndefinedAttribute(name, dir(_object))


_resolve_udn = _resolve_udn_prefer_attr3


def _resolve_placeholder(name, template, global_vars):
    placeholder_cache = template.placeholder_cache
    if placeholder_cache and name in placeholder_cache:
        ph = placeholder_cache[name]
        if isinstance(ph, weakref.ReferenceType):
            v = ph()
            if v is not None:
                return v
        else:
            return ph

    # Note: getattr with 3 args is somewhat slower if the attribute
    # is found, but much faster if the attribute is not found.
    udn_ph = UndefinedPlaceholder
    result = getattr(template, name, udn_ph)
    if result is not udn_ph:
        if placeholder_cache is not None:
            # Use a weakref for methods to prevent memory cycles.
            placeholder_cache[name] = weakref.ref(result) if inspect.ismethod(
                result) else result
        return result

    search_list = template.search_list
    if search_list:
        ph = resolve_from_search_list(search_list, name)
        if ph is not UnresolvedPlaceholder:
            if placeholder_cache is not None:
                # Use a weakref for methods to prevent memory cycles.
                placeholder_cache[name] = weakref.ref(ph) if inspect.ismethod(
                    ph) else ph
            return ph

    # TODO: Cache negative results in placedholder_cache?
    # This probably isn't worthwhile as it likely won't happen often enough
    # to make the extra code/cpu/memory worthwhile.

    if global_vars is not None:
        try:
            return global_vars[name]
        except KeyError:
            pass
        except TypeError:
            raise runtime.PlaceholderError(
                'unexpected type for global_vars: %s' % type(global_vars))

    # fixme: finally try to resolve builtins - this should be configurable
    # if you compile optimized modes, this isn't necessary
    try:
        return getattr(__builtins__, name)
    except AttributeError:
        return UndefinedPlaceholder(name, search_list)


# FIXME: i'm sure this is a little pokey - might be able to speed this up
# somehow. not sure if it's better to look before leaping or raise.
# might also want to let users tune whether to prefer keys or attributes
def _resolve_placeholder_with_locals(name, template, local_vars, global_vars):
    if local_vars is not None:
        try:
            return local_vars[name]
        except KeyError:
            pass
        except TypeError:
            raise runtime.PlaceholderError(
                'unexpected type for local_vars: %s' % type(local_vars))

    return _resolve_placeholder(name, template, global_vars)


def _debug_resolve_placeholder(name, *pargs, **kargs):
    placeholder = _resolve_placeholder(name, *pargs, **kargs)
    if inspect.isroutine(placeholder):
        return CallOnlyPlaceholder(name, placeholder)
    else:
        return placeholder


def _debug_resolve_udn(_object, name, *pargs, **kargs):
    placeholder = _resolve_udn_prefer_attr3(_object, name, *pargs, **kargs)
    if inspect.isroutine(placeholder):
        return CallOnlyPlaceholder(name, placeholder)
    else:
        return placeholder


def _resolve_from_search_list(search_list, name, default=Unspecified):
    try:
        for scope in search_list:
            try:
                return scope[name]
            except (TypeError, KeyError):
                pass

            try:
                return getattr(scope, name)
            except AttributeError:
                pass
    except TypeError:
        # if this isn't iterable, let's just return UndefinedPlaceholder
        pass

    if default != Unspecified:
        return default
    else:
        return UnresolvedPlaceholder

# Define Python/C alternates.
_python_resolve_from_search_list = _resolve_from_search_list
_python_resolve_udn = _resolve_udn
if _udn:
    _c_resolve_from_search_list = _udn._resolve_from_search_list
    _c_resolve_udn = _udn._resolve_udn

# Set default functions.
resolve_from_search_list = _python_resolve_from_search_list
resolve_udn = _python_resolve_udn
resolve_placeholder = _resolve_placeholder
resolve_placeholder_with_locals = _resolve_placeholder_with_locals


def set_accelerator(enabled=True, enable_test_mode=False):
    """Some key functions are much faster in C.
    They can subtlely change how data is accessed which can cause false-positive
    errors in certain test cases, so we want to be able to toggle it on/off.
    """
    global resolve_from_search_list
    global resolve_udn

    if enabled and _udn:
        resolve_from_search_list = _c_resolve_from_search_list
        resolve_udn = _c_resolve_udn
    elif enable_test_mode:
        resolve_from_search_list = _python_resolve_from_search_list
        # use this resolver so that we don't call resolve tester attributes twice
        # automatically - this screws up testing hoisting and other things that
        # are designed to limit calls to resolve_placeholder
        resolve_udn = _resolve_udn_prefer_attr
    else:
        resolve_from_search_list = _python_resolve_from_search_list
        resolve_udn = _python_resolve_udn

    if enabled and _udn is None:
        logging.warning('unable to enable acceleration, _udn module not loaded')

# give it our best shot
set_accelerator()



# Copyright 2015 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import timeit

from spitfire.runtime import template
from spitfire.runtime import baked


def skip():
    pass


skip.skip_filter = True


def no_skip():
    pass


sp_creation = ['foo', 'adafadfhasdkfh',
               'dsafjadhf dafhaskfh kadsfadfadfasfjakfh',
               'sdfsdfaa  adksj k jasdfljkasdf jka dsf']
mark_as_s = ['foo', object(), 12344345234274.3333,
             'asdfasdf asdjfh asjfd asdfasdf', ['123']]


def time_fn(label, fn, arg):
    if hasattr(arg, '__len__') and len(arg) == 2:
        v, f = arg
        print('Timing %s - %s(%s, %s)' % (label, fn.__name__, v, f.__name__))
        t = timeit.Timer(lambda: fn(v, f))
    else:
        print('Timing %s - %s(%s)' % (label, fn.__name__, arg))
        t = timeit.Timer(lambda: fn(arg))
    try:
        r = t.repeat(10, 100000)
    except:  # pylint:disable=bare-except
        t.print_exc()
        return

    best = min(r)
    scale = 1e3 / len(r)
    msec = best * scale
    all = ', '.join('%.1f' % (v * scale) for v in r)
    print('best run: %.1f msec per loop [%s]' % (msec, all))


def main():
    c_tmpl = template.get_spitfire_template_class(prefer_c_extension=True)()
    py_tmpl = template.get_spitfire_template_class(prefer_c_extension=False)()

    print('*** filter_function with SanitizedPlaceholder')
    for v in sp_creation:
        sp = baked.SanitizedPlaceholder(v)
        time_fn('Py', py_tmpl.filter_function, sp)
        time_fn('C', c_tmpl.filter_function, sp)
        print('')

    print('***filter_function with string')
    for v in sp_creation:
        time_fn('Py', py_tmpl.filter_function, v)
        time_fn('C', c_tmpl.filter_function, v)
        print('')

    print('***filter_function with skip_filter')
    for v in sp_creation:
        time_fn('Py', py_tmpl.filter_function, (v, skip))
        time_fn('C', c_tmpl.filter_function, (v, skip))
        print('')

    print('***filter_function with no skip_filter')
    for v in sp_creation:
        time_fn('Py', py_tmpl.filter_function, (v, no_skip))
        time_fn('C', c_tmpl.filter_function, (v, no_skip))
        print('')


if __name__ == '__main__':
    main()



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import logging

import io as StringIO

from spitfire.compiler import ast


class CodegenError(Exception):
    pass


class CodeNode(object):

    def __init__(self, src_line=None, input_pos=None):
        self.src_line = src_line
        self.input_pos = input_pos
        self.child_nodes = []

    def append_line(self, line):
        self.append(CodeNode(line))

    def append(self, code_node):
        self.child_nodes.append(code_node)

    def insert(self, index, code_node):
        self.child_nodes.insert(index, code_node)

    def extend(self, code_nodes):
        try:
            self.child_nodes.extend(code_nodes)
        except TypeError:
            raise CodegenError("can't add %s" % code_nodes)

    def __repr__(self):
        return '%s:%s' % (self.__class__.__name__, self.src_line)


# perform an in-order traversal of the AST and call the generate methods
# in this case, we are generating python source code that should be somewhat
# human-readable
class CodeGenerator(object):
    indent_str = '  '
    indent_level = 0

    # options - an AnalyzerOptions object
    def __init__(self, ast_root, compiler, options=None):
        self.ast_root = ast_root
        # Compiler - an instance of Compiler defined in util.
        self.compiler = compiler
        # This stack of FunctionNodes represents the functions we are
        # currently inside. When we enter a function node, we push that
        # onto the stack and when we leave that function, we pop it from
        # the stack.
        self.function_stack = []
        self.options = options
        self.output = StringIO.StringIO()
        self.template = None
        self.baked_mode = False

    def get_code(self):
        code_root = self.build_code(self.ast_root)[0]
        self.write_python(code_root, indent_level=-1)
        return self.output.getvalue().encode(self.ast_root.encoding)

    def generate_python(self, code_node):
        try:
            return code_node.src_line
        except AttributeError as e:
            self.compiler.error(CodegenError("can't write code_node: %s\n\t%s" %
                                             (code_node, e)))

    def write_python(self, code_node, indent_level):
        try:
            if code_node.src_line is not None:
                if code_node.src_line:
                    self.output.write(self.indent_str * indent_level)
                    self.output.write(code_node.src_line)
                    if (self.options.include_sourcemap and
                            code_node.input_pos and self.compiler.src_line_map):
                        self.output.write(
                            ' # L%s' %
                            self.compiler.src_line_map[code_node.input_pos])
                self.output.write('\n')
        except AttributeError:
            self.compiler.error(CodegenError("can't write code_node: %s" %
                                             code_node))

        for cn in code_node.child_nodes:
            self.write_python(cn, indent_level + 1)

    def build_code(self, ast_node):
        method_name = 'codegenAST%s' % ast_node.__class__.__name__
        method = getattr(self, method_name, self.codegenDefault)
        return method(ast_node)

    def codegenASTTemplateNode(self, node):
        self.template = node.copy(copy_children=False)

        module_code = CodeNode()
        module_code.append_line('#!/usr/bin/env python')
        module_code.append_line('# -*- coding: %s -*-' % node.encoding)
        if node.baked:
            module_code.append_line('# Baked Mode')
            self.baked_mode = True
        module_code.append_line('')
        if node.import_nodes:
            module_code.append_line('# template imports')
            for n in node.import_nodes:
                module_code.extend(self.build_code(n))
            module_code.append_line('')
        if node.from_nodes:
            module_code.append_line('# template from imports')
            for n in node.from_nodes:
                module_code.extend(self.build_code(n))
            module_code.append_line('')

        classname = node.classname

        module_code.append_line('import spitfire.runtime')
        module_code.append_line('import spitfire.runtime.template')

        if self.options and self.options.cheetah_cheats:
            module_code.append_line(
                'from Cheetah.NameMapper import valueFromSearchList '
                'as resolve_placeholder')
            module_code.append_line(
                'from Cheetah.NameMapper import valueForKey as resolve_udn')
        else:
            module_code.append_line(
                'from spitfire.runtime.udn import resolve_placeholder')
            module_code.append_line('from spitfire.runtime.udn '
                                    'import resolve_placeholder_with_locals')
            module_code.append_line(
                'from spitfire.runtime.udn import resolve_udn')

        module_code.append_line(
            'from spitfire.runtime.baked import SanitizedPlaceholder')
        module_code.append_line(
            'from spitfire.runtime.baked import runtime_mark_as_sanitized')
        module_code.append_line(
            'from spitfire.runtime.baked import mark_as_sanitized')
        module_code.append_line(
            'from spitfire.runtime.template import template_method')
        module_code.append_line('')

        if node.cached_identifiers:
            module_code.append_line('# cached identifiers')
            for cached_ph in node.cached_identifiers:
                module_code.append_line('%s = None' % cached_ph.name)
            module_code.append_line('')

        if not node.library:
            extends = []
            for n in node.extends_nodes:
                extends.append(self.generate_python(self.build_code(n)[0]))
            if not extends:
                extends = [self.options.base_template_full_import_path]
                if (self.options.base_template_full_import_path !=
                    (self.options.DEFAULT_BASE_TEMPLATE_FULL_IMPORT_PATH)):
                    # If we are not using the default base template class, we
                    # need to import the module for the custom template class.
                    # The module for the default base template class is already
                    # imported below and is required regardless of which base
                    # template is being used.
                    base_template_module_bits = (
                        self.options.base_template_full_import_path.split(
                            '.')[:-1])
                    # If the full import path is a.b.ClassName, we want to
                    # import the module a.b.
                    base_template_module = '.'.join(base_template_module_bits)
                    module_code.append_line('import %s' % base_template_module)

            extends_clause = ', '.join(extends)
            class_code = CodeNode('class %(classname)s(%(extends_clause)s):' %
                                  vars())

            module_code.append(class_code)
            for n in node.attr_nodes:
                class_code.extend(self.build_code(n))
                class_code.append_line('')
            def_parent_code = class_code
        else:
            # Library functions are written to the module directly.
            module_code.append_line('')
            def_parent_code = module_code

        for n in node.child_nodes:
            def_parent_code.extend(self.build_code(n))
            def_parent_code.append_line('')

        # if we aren't extending a template, build out the main function
        if not node.library and (not node.extends_nodes or node.implements):
            def_parent_code.extend(self.build_code(node.main_function))

        # NOTE(msolo): originally, i thought this would be helpful in case a bit
        # of human error - however, a more robust check is necessary here to
        # make the warning less spurious
        # else:
        #   from spitfire.compiler.visitor import flatten_tree
        #   logging.warning(throwing away defined main function because "
        #                   "it is not a base class %s %s",
        #                   self.ast_root.source_path)
        #   logging.warning("%s", flatten_tree(node.main_function))

        module_code.append_line(run_tmpl % vars(node))

        return [module_code]

    def codegenASTExtendsNode(self, node):
        return [CodeNode('.'.join([
            self.generate_python(self.build_code(n)[
                0]) for n in node.module_name_list
        ]))]

    def codegenASTImportNode(self, node):
        return [CodeNode('import %s' % '.'.join([
            self.generate_python(self.build_code(n)[
                0]) for n in node.module_name_list
        ]))]

    def codegenASTFromNode(self, node):
        from_clause = '.'.join([
            self.generate_python(self.build_code(n)[
                0]) for n in node.module_name_list
        ])
        import_clause = self.generate_python(self.build_code(node.identifier)[
            0])

        if node.alias:
            alias_clause = self.generate_python(self.build_code(node.alias)[0])
            return [CodeNode('from %(from_clause)s '
                             'import %(import_clause)s '
                             'as %(alias_clause)s' % vars())]
        else:
            return [CodeNode('from %(from_clause)s '
                             'import %(import_clause)s' % vars())]

    def codegenASTPlaceholderSubstitutionNode(self, node):
        placeholder = self.generate_python(self.build_code(node.expression)[0])
        return [CodeNode(ASTPlaceholderSubstitutionNode_tmpl[0] % vars(),
                         input_pos=node.pos)]

    def codegenASTDoNode(self, node):
        return [CodeNode(
            self.generate_python(self.build_code(node.expression)[0]),
            input_pos=node.pos)]

    def codegenASTCallFunctionNode(self, node):
        expression = self.generate_python(self.build_code(node.expression)[0])
        if expression == '_self_filter_function':
            self.function_stack[-1].uses_filter_function = True
        if expression == '_self_private_filter_function':
            self.function_stack[-1].uses_private_filter_function = True
        if node.arg_list:
            arg_list = self.generate_python(self.build_code(node.arg_list)[0])
        else:
            arg_list = ''
        call = ASTCallFunctionNode_tmpl[0] % vars()
        if self.baked_mode:
            sanitization_state = node.sanitization_state
            # For SANITIZED_STRING we could use SanitizedPlaceholder(), but the
            # mark_as_sanitized() function is faster so we use that instead.
            if (sanitization_state == ast.SanitizedState.SANITIZED_STRING or
                    sanitization_state == ast.SanitizedState.SANITIZED):
                return [CodeNode('mark_as_sanitized(%s)' % call,
                                 input_pos=node.pos)]
            elif (sanitization_state == ast.SanitizedState.UNSANITIZED or
                  sanitization_state == ast.SanitizedState.NOT_OUTPUTTED or
                  sanitization_state ==
                  ast.SanitizedState.OUTPUTTED_IMMEDIATELY):
                return [CodeNode(call, input_pos=node.pos)]
            elif sanitization_state == ast.SanitizedState.UNKNOWN:
                return [CodeNode(
                    'runtime_mark_as_sanitized(%(call)s, %(expression)s)' %
                    vars(),
                    input_pos=node.pos)]
        return [CodeNode(call, input_pos=node.pos)]

    def codegenASTForNode(self, node):
        target_list = self.generate_python(self.build_code(node.target_list)[0])
        expression_list = self.generate_python(self.build_code(
            node.expression_list)[0])
        code_node = CodeNode(ASTForNode_tmpl[0] % vars(), input_pos=node.pos)
        for n in node.child_nodes:
            code_node.extend(self.build_code(n))
        return [code_node]

    def codegenASTIfNode(self, node):
        test_expression = self.generate_python(self.build_code(
            node.test_expression)[0])
        if_code_node = CodeNode("if %(test_expression)s:" % vars(),
                                input_pos=node.pos)
        for n in node.child_nodes:
            if_code_node.extend(self.build_code(n))
        code_nodes = [if_code_node]
        if node.else_.child_nodes:
            else_code_node = CodeNode('else:')
            for n in node.else_.child_nodes:
                else_code_node.extend(self.build_code(n))
            code_nodes.append(else_code_node)
        return code_nodes

    def codegenASTTargetListNode(self, node):
        if len(node.child_nodes) == 1:
            return self.build_code(node.child_nodes[0])
        else:
            return [CodeNode('(%s)' % ', '.join([self.generate_python(
                self.build_code(n)[0]) for n in node.child_nodes]))]

    codegenASTExpressionListNode = codegenASTTargetListNode

    def codegenASTLiteralNode(self, node):
        if (self.options and not self.options.generate_unicode and
                isinstance(node.value, str)):
            # If the node is the empty string, we should mark it as sanitized by
            # default. Eventually, all string literals should be marked as
            # sanitized.
            if self.baked_mode and node.value == '':
                return [CodeNode("mark_as_sanitized('')", input_pos=node.pos)]

            return [CodeNode(
                repr(node.value.encode(self.ast_root.encoding)),
                input_pos=node.pos)]
        else:
            # generate unicode by default
            return [CodeNode('%(value)r' % vars(node), input_pos=node.pos)]

    def codegenASTListLiteralNode(self, node):
        return [CodeNode('[%s]' % ', '.join([
            self.generate_python(self.build_code(n)[
                0]) for n in node.child_nodes
        ]),
                         input_pos=node.pos)]

    def codegenASTTupleLiteralNode(self, node):
        return [CodeNode('(%s)' % ', '.join([
            self.generate_python(self.build_code(n)[
                0]) for n in node.child_nodes
        ]),
                         input_pos=node.pos)]

    def codegenASTDictLiteralNode(self, node):
        return [
            CodeNode('{%s}' % ', '.join([
                '%s: %s' % (self.generate_python(self.build_code(kn)[
                    0]), self.generate_python(self.build_code(vn)[0])
                           ) for kn, vn in node.child_nodes
            ]))
        ]

    def codegenASTParameterNode(self, node):
        if node.default:
            return [CodeNode('%s=%s' % (
                node.name,
                self.generate_python(self.build_code(node.default)[0])))]
        else:
            return [CodeNode('%s' % node.name)]

    def codegenASTAttributeNode(self, node):
        return [CodeNode('%s = %s' % (
            node.name, self.generate_python(self.build_code(node.default)[0])),
                         input_pos=node.pos)]

    def codegenASTFilterAttributeNode(self, node):
        return [CodeNode('%s = staticmethod(%s)' % (
            node.name, self.generate_python(self.build_code(node.default)[0])))]

    def codegenASTParameterListNode(self, node):
        if len(node.child_nodes) == 1:
            return self.build_code(node.child_nodes[0])
        else:
            return [CodeNode('%s' % ', '.join([self.generate_python(
                self.build_code(n)[0]) for n in node.child_nodes]))]

    codegenASTArgListNode = codegenASTParameterListNode

    def codegenASTGetUDNNode(self, node):
        #print ' '.join("codegenASTGetUDNNode", id(node), "name", node.name,
        #               "expr", node.expression)
        expression = self.generate_python(self.build_code(node.expression)[0])
        name = node.name
        if (self.options and self.options.default_to_strict_resolution and
                self.template and not self.template.use_loose_resolution):
            return [CodeNode("%(expression)s.%(name)s" % vars())]
        if self.options and self.options.raise_udn_exceptions:
            return [CodeNode(
                "resolve_udn(%(expression)s, '%(name)s', raise_exception=True)"
                % vars(),
                input_pos=node.pos)]
        else:
            return [CodeNode("resolve_udn(%(expression)s, '%(name)s')" % vars(),
                             input_pos=node.pos)]

    def codegenASTPlaceholderNode(self, node):
        name = node.name
        if name in ('has_var', 'get_var'):
            return [CodeNode("self.%(name)s" % vars())]
        elif self.options and self.options.cheetah_cheats:
            return [CodeNode("resolve_placeholder([locals()] + "
                             "_self_search_list, '%(name)s')" % vars(),
                             input_pos=node.pos)]
        elif self.options and self.options.omit_local_scope_search:
            self.function_stack[-1].uses_globals = True
            return [CodeNode("resolve_placeholder('%(name)s', self, _globals)" %
                             vars(),
                             input_pos=node.pos)]
        else:
            self.function_stack[-1].uses_globals = True
            return [CodeNode("resolve_placeholder_with_locals('%(name)s', "
                             "self, locals(), _globals)" % vars(),
                             input_pos=node.pos)]

    def codegenASTReturnNode(self, node):
        expression = self.generate_python(self.build_code(node.expression)[0])
        return [CodeNode("return %(expression)s" % vars(), input_pos=node.pos)]

    def codegenASTOptionalWhitespaceNode(self, node):
        #if self.ignore_optional_whitespace:
        #  return []
        return [CodeNode(ASTOptionalWhitespaceNode_tmpl[0] % vars(node),
                         input_pos=node.pos)]

    def codegenASTSliceNode(self, node):
        expression = self.generate_python(self.build_code(node.expression)[0])
        slice_expression = self.generate_python(self.build_code(
            node.slice_expression)[0])
        return [CodeNode("%(expression)s[%(slice_expression)s]" % vars())]

    def codegenASTBinOpExpressionNode(self, node):
        left = self.generate_python(self.build_code(node.left)[0])
        right = self.generate_python(self.build_code(node.right)[0])
        operator = node.operator
        return [CodeNode('(%(left)s %(operator)s %(right)s)' % vars(),
                         input_pos=node.pos)]

    def codegenASTBinOpNode(self, node):
        left = self.generate_python(self.build_code(node.left)[0])
        right = self.generate_python(self.build_code(node.right)[0])
        operator = node.operator
        return [CodeNode('%(left)s %(operator)s %(right)s' % vars(),
                         input_pos=node.pos)]

    codegenASTAssignNode = codegenASTBinOpNode

    def codegenASTUnaryOpNode(self, node):
        expression = self.generate_python(self.build_code(node.expression)[0])
        operator = node.operator
        return [CodeNode('(%(operator)s %(expression)s)' % vars(),
                         input_pos=node.pos)]

    def codegenASTGetAttrNode(self, node):
        expression = self.generate_python(self.build_code(node.expression)[0])
        name = node.name
        return [CodeNode("%(expression)s.%(name)s" % vars(),
                         input_pos=node.pos)]

    def codegenASTFunctionNode(self, node):
        name = node.name
        node.uses_globals = False
        node.uses_filter_function = False
        node.uses_private_filter_function = False
        node.uses_buffer_write = False
        node.uses_buffer_extend = False
        self.function_stack.append(node)
        if node.parameter_list:
            parameter_list = self.generate_python(self.build_code(
                node.parameter_list)[0])
        else:
            parameter_list = ''

        decorator_node = CodeNode('@template_method')
        # NOTE: for Cheetah compatibility, we have to handle the case where
        # Cheetah tries to pass a 'transaction' object through. hopefully this
        # doesn't have some other baggage coming with it.
        if self.options and self.options.cheetah_compatibility:
            if parameter_list:
                code_node = CodeNode(
                    'def %(name)s(%(parameter_list)s, **kargs):' % vars(),
                    input_pos=node.pos)
            else:
                code_node = CodeNode('def %(name)s(**kargs):' % vars(),
                                     input_pos=node.pos)
        else:
            code_node = CodeNode('def %(name)s(%(parameter_list)s):' % vars(),
                                 input_pos=node.pos)

        needs_globals_added = True
        child_nodes = node.child_nodes

        if self.options and self.options.cheetah_compatibility:
            if_cheetah = CodeNode("if 'trans' in kargs:")
            code_node.append(if_cheetah)
            if_cheetah.append(CodeNode("_buffer = kargs['trans'].response()"))
            else_spitfire = CodeNode('else:')
            else_spitfire.append(CodeNode('_buffer = self.new_buffer()'))
            code_node.append(else_spitfire)
        else:
            # If the first node is an if statement with no else AND there's no
            # other statements, generate the if code first.  That way, we can
            # avoid doing extra work when the condition is false.  ie, avoid the
            # overhead of creating a new list setting up useless local variables
            # and joining all to get an empty string. Disable this in baked mode
            # until I figure out how to handle this. TODO: Do not perform
            # sanitization inside of a test_expression. Then we can remove this
            # baked_mode check.
            if child_nodes and len(child_nodes) == 1 and not self.baked_mode:
                if_node = child_nodes[0]
                if isinstance(if_node,
                              ast.IfNode) and not if_node.else_.child_nodes:
                    child_nodes = if_node.child_nodes
                    # Insert code that does:
                    #   if not ($test_expression):
                    #     return ''
                    new_if_condition = ast.IfNode(ast.UnaryOpNode(
                        'not', if_node.test_expression))
                    new_if_condition.append(ast.ReturnNode(ast.LiteralNode('')))
                    new_code = self.build_code(new_if_condition)
                    if node.uses_globals:
                        needs_globals_added = False
                        code_node.append(CodeNode('_globals = globals()'))
                    code_node.extend(new_code)

            code_node.append(CodeNode('_buffer = self.new_buffer()'))

        # Save the point where _globals and self_filter_funtion will go if used.
        # We don't append these here because we have to determine if these two
        # functions are used in the scope of the current ast.FunctionNode.
        insertion_point = len(code_node.child_nodes)

        if self.options and self.options.cheetah_cheats:
            node.uses_globals = True
            if_cheetah = CodeNode('if self.search_list:')
            if_cheetah.append(CodeNode(
                '_self_search_list = self.search_list + [_globals]'))
            else_cheetah = CodeNode('else:')
            else_cheetah.append(CodeNode('_self_search_list = [_globals]'))
            code_node.append(if_cheetah)
            code_node.append(else_cheetah)

        for n in child_nodes:
            code_child_nodes = self.build_code(n)
            code_node.extend(code_child_nodes)

        if node.uses_globals and needs_globals_added:
            code_node.insert(insertion_point, CodeNode('_globals = globals()'))
            insertion_point += 1
        if node.uses_filter_function:
            code_node.insert(
                insertion_point,
                CodeNode('_self_filter_function = self.filter_function'))
            insertion_point += 1
        if node.uses_private_filter_function:
            code_node.insert(insertion_point, CodeNode(
                '_self_private_filter_function = self._filter_function'))
            insertion_point += 1
        if node.uses_buffer_write:
            code_node.insert(insertion_point,
                             CodeNode('_buffer_write = _buffer.write'))
            insertion_point += 1
        if node.uses_buffer_extend:
            code_node.insert(insertion_point,
                             CodeNode('_buffer_extend = _buffer.extend'))
        if self.options.cheetah_compatibility:
            if_cheetah = CodeNode("if 'trans' not in kargs:")
            if_cheetah.append(CodeNode('return _buffer.getvalue()'))
            code_node.append(if_cheetah)
        else:
            code_node.append(CodeNode('return _buffer.getvalue()'))
        self.function_stack.pop()
        return [decorator_node, code_node]

    # fixme: don't know if i still need this - a 'template function'
    # has an implicit return of the buffer built in - might be simpler
    # to code that rather than adding a return node during the analyze
    #def codegenASTReturnNode(self, node):
    #  code_node = self.codegenDefault(node)

    def codegenASTBufferWrite(self, node):
        self.function_stack[-1].uses_buffer_write = True
        expression = self.generate_python(self.build_code(node.expression)[0])
        code_node = CodeNode('_buffer_write(%(expression)s)' % vars(),
                             input_pos=node.pos)
        return [code_node]

    def codegenASTBufferExtend(self, node):
        self.function_stack[-1].uses_buffer_extend = True
        expression = self.generate_python(self.build_code(node.expression)[0])
        code_node = CodeNode('_buffer_extend(%(expression)s)' % vars(),
                             input_pos=node.pos)
        return [code_node]

    def codegenASTEchoNode(self, node):
        self.function_stack[-1].uses_buffer_write = True
        node_list = []

        true_expression = self.generate_python(self.build_code(
            node.true_expression)[0])
        true_code = CodeNode('_buffer_write(%(true_expression)s)' % vars())
        if node.test_expression:
            test_expression = self.generate_python(self.build_code(
                node.test_expression)[0])
            if_code = CodeNode('if %(test_expression)s:' % vars())
            if_code.append(true_code)
            node_list.append(if_code)
        else:
            node_list.append(true_code)

        if node.false_expression:
            false_expression = self.generate_python(self.build_code(
                node.false_expression)[0])
            else_code = CodeNode('else:' % vars())
            else_code.append(CodeNode('_buffer_write(%(false_expression)s)' %
                                      vars()))
            node_list.append(else_code)
        return node_list

    def codegenASTCacheNode(self, node):
        self.function_stack[-1].uses_globals = True
        cached_name = node.name
        expression = self.generate_python(self.build_code(node.expression)[0])
        # use dictionary syntax to get around coalescing 'global' statements
        #globalize_var = CodeNode('global %(cached_name)s' % vars())
        if_code = CodeNode("if %(cached_name)s is None:" % vars(),
                           input_pos=node.pos)
        if_code.append(
            CodeNode("_globals['%(cached_name)s'] = %(expression)s" % vars(),
                     input_pos=node.pos))
        return [if_code]

    def codegenASTFilterNode(self, node):
        expression = self.generate_python(self.build_code(node.expression)[0])
        if node.filter_function_node == ast.DefaultFilterFunction:
            # In baked mode, we must always call into _self_filter_function.
            # This is because this is the function that will contain the logic
            # for deciding whether or not to filter a SanitizedPlaceholder.
            if isinstance(node.expression,
                          ast.CallFunctionNode) or self.baked_mode:
                self.function_stack[-1].uses_filter_function = True
                filter_expression = '_self_filter_function'
            else:
                # Since there is no second argument to filter_function, there
                # will be a call to _filter_function. This optimization skips
                # the check and calls directly to _filter_function.
                self.function_stack[-1].uses_private_filter_function = True
                filter_expression = '_self_private_filter_function'
        elif node.filter_function_node:
            filter_expression = self.generate_python(self.build_code(
                node.filter_function_node)[0])
        else:
            filter_expression = None

        if isinstance(node.expression, ast.CallFunctionNode):
            # need the placeholder function expression to make sure that we
            # don't double escape the output of template functions
            # fixme: this is suboptimal if this expression is expensive -
            # should the optimizer fix this, or should we generate speedy code?
            placeholder_function_expression = self.generate_python(
                self.build_code(node.expression.expression)[0])
            if node.filter_function_node == ast.DefaultFilterFunction:
                code_node = CodeNode('%(filter_expression)s(%(expression)s, '
                                     '%(placeholder_function_expression)s)' %
                                     vars(),
                                     input_pos=node.pos)
            elif node.filter_function_node:
                code_node = CodeNode(
                    '%(filter_expression)s(self, '
                    '%(expression)s, %(placeholder_function_expression)s)' %
                    vars(),
                    input_pos=node.pos)
            else:
                code_node = CodeNode('%(expression)s' % vars(),
                                     input_pos=node.pos)
        else:
            if node.filter_function_node == ast.DefaultFilterFunction:
                code_node = CodeNode('%(filter_expression)s(%(expression)s)' %
                                     vars(),
                                     input_pos=node.pos)
            elif node.filter_function_node:
                code_node = CodeNode('%(filter_expression)s(%(expression)s)' %
                                     vars(),
                                     input_pos=node.pos)
            else:
                code_node = CodeNode('%(expression)s' % vars(),
                                     input_pos=node.pos)
        return [code_node]

    def codegenDefault(self, node):
        v = globals()
        try:
            return [CodeNode(line % vars(node))
                    for line in v['AST%s_tmpl' % node.__class__.__name__]]
        except KeyError as e:
            self.compiler.error(CodegenError("no codegen for %s %s" % (type(
                node), vars(node))))

    def codegen(self, node):
        return self.codegenDefault(node)[0]


run_tmpl = """

if __name__ == '__main__':
  import spitfire.runtime.runner
  spitfire.runtime.runner.run_template(%(classname)s)
"""

ASTFunctionNode_tmpl = ['def %(name)s(%(parameter_list)s):']

ASTCallFunctionNode_tmpl = ['%(expression)s(%(arg_list)s)']

ASTForNode_tmpl = ['for %(target_list)s in %(expression_list)s:']

ASTTargetNode_tmpl = ['%(name)s']

ASTIdentifierNode_tmpl = ['%(name)s']
ASTTemplateMethodIdentifierNode_tmpl = ['self.%(name)s']

ASTLiteralNode_tmpl = ['%(value)r']

ASTBreakNode_tmpl = ['break']

ASTContinueNode_tmpl = ['continue']



# Copyright 2014 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import unittest
import weakref

from spitfire import runtime
from spitfire.runtime import udn
try:
    from spitfire.runtime import _udn  # pylint: disable=g-import-not-at-top
except ImportError:
    _udn = None


class Scope(object):
    boom = 'bam'


class Foo(object):
    bar = 'baz'
    search_list = [{'win': 'boo'}, Scope(),]
    placeholder_cache = None

    def foo_method(self):
      pass


class TestResolvePlaceholder(unittest.TestCase):

    def test_has_attr(self):
        self.assertEqual(udn.resolve_placeholder('bar', Foo, None), 'baz')

    def test_in_search_list_dict(self):
        self.assertEqual(udn.resolve_placeholder('win', Foo, None), 'boo')

    def test_in_search_list_object(self):
        self.assertEqual(udn.resolve_placeholder('boom', Foo, None), 'bam')

    def test_in_globals(self):
        self.assertEqual(
            udn.resolve_placeholder('blam', Foo, {'blam': 'bling'}), 'bling')

    def test_builtin(self):
        self.assertEqual(udn.resolve_placeholder('str', Foo, None), str)

    def test_undefined(self):
        self.assertEqual(
            type(udn.resolve_placeholder('wowza', Foo, None)),
            runtime.UndefinedPlaceholder)


class TestResolvePlaceholderWithCache(unittest.TestCase):

    def test_in_search_list_dict(self):
        template = Foo()
        template.placeholder_cache = {}
        self.assertEqual(udn.resolve_placeholder('win', template, None), 'boo')
        self.assertIn('win', template.placeholder_cache)
        self.assertEqual(udn.resolve_placeholder('win', template, None), 'boo')

    def test_in_search_list_object(self):
        template = Foo()
        template.placeholder_cache = {}
        self.assertEqual(udn.resolve_placeholder('boom', template, None), 'bam')
        self.assertIn('boom', template.placeholder_cache)
        self.assertEqual(udn.resolve_placeholder('boom', template, None), 'bam')

    def test_method_cycle_elimination(self):
        template = Foo()
        template.placeholder_cache = {}
        self.assertEqual(udn.resolve_placeholder('foo_method', template, None),
                         template.foo_method)
        self.assertIn('foo_method', template.placeholder_cache)
        self.assertIsInstance(template.placeholder_cache['foo_method'],
                              weakref.ReferenceType)
        self.assertEqual(udn.resolve_placeholder('foo_method', template, None),
                         template.foo_method)


class TestResolvePlaceholderWithLocals(unittest.TestCase):

    def test_in_locals(self):
        self.assertEqual(
            udn.resolve_placeholder_with_locals('foo', Foo, {'foo': 'bar'},
                                                None), 'bar')


class Baz(object):
    bar = 'win'


class _UdnTest(object):

    resolve_udn = None

    def testResolveHitWithoutException(self):
        self.assertEqual(self.resolve_udn(Baz(), 'bar'), 'win')

    def testResolveHitWithException(self):
        self.assertEqual(
            self.resolve_udn(Baz(),
                             'bar',
                             raise_exception=True),
            'win')

    def testResolveMissWithoutException(self):
        self.assertIsInstance(
            self.resolve_udn(Baz(), 'missing'), runtime.UndefinedAttribute)

    def testResolveDoubleMissWithoutException(self):
        """Shows that it's UndefinedAttribute's all the way down."""
        undefined_attr = self.resolve_udn(Baz(), 'missing')
        self.assertIsInstance(undefined_attr, runtime.UndefinedAttribute)
        undefined_attr2 = self.resolve_udn(undefined_attr, 'missing')
        self.assertIsInstance(undefined_attr2, runtime.UndefinedAttribute)

    def testResolveMissWithException(self):
        self.assertRaises(runtime.UDNResolveError,
                          self.resolve_udn,
                          Baz(),
                          'misssing',
                          raise_exception=True)


if _udn is not None:

    class TestUdnC(_UdnTest, unittest.TestCase):
        resolve_udn = staticmethod(_udn._resolve_udn)


class TestUdnPyAttr3(_UdnTest, unittest.TestCase):
    resolve_udn = staticmethod(udn._resolve_udn_prefer_attr3)


class TestUdnPyAttr2(_UdnTest, unittest.TestCase):
    resolve_udn = staticmethod(udn._resolve_udn_prefer_attr2)


class TestUdnPyAttr(_UdnTest, unittest.TestCase):
    resolve_udn = staticmethod(udn._resolve_udn_prefer_attr)


class TestUdnPyDict(_UdnTest, unittest.TestCase):
    resolve_udn = staticmethod(udn._resolve_udn_prefer_dict)


if __name__ == '__main__':
    unittest.main()



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import io


class TreeWalkError(Exception):
    pass


class VisitNode(object):

    def __init__(self, node_repr=None, child_nodes=None):
        self.node_repr = node_repr
        if child_nodes is None:
            self.child_nodes = []
        else:
            self.child_nodes = child_nodes

    def append_line(self, line):
        self.append(VisitNode(line))

    def append(self, visit_node):
        self.child_nodes.append(visit_node)

    def extend(self, visit_nodes):
        try:
            self.child_nodes.extend(visit_nodes)
        except TypeError:
            raise TreeWalkError("can't add %s" % visit_nodes)

    def __repr__(self):
        return '%s:%s' % (self.__class__.__name__, self.node_repr)


def flatten_tree(root):
    return TreeVisitor(root).get_text()


def print_tree(root, output=None):
    if output:
        print(flatten_tree(root), file=output)
    else:
        print(flatten_tree(root))


# perform an in-order traversal of the AST and call the generate methods
# in this case, we are generating python source code that should be somewhat
# human-readable
class TreeVisitor(object):
    indent_str = '  '
    indent_level = 0

    def __init__(self, ast_root):
        self.ast_root = ast_root
        self.output = io.StringIO()

    def get_text(self):
        root = self.build_text(self.ast_root)[0]
        self.write_visit(root)
        text = self.output.getvalue()
        try:
            text = text.encode(self.ast_root.encoding)
        except AttributeError as e:
            pass
        return text

    def generate_text(self, visit_node):
        try:
            return visit_node.node_repr
        except AttributeError as e:
            raise TreeWalkError("can't write visit_node: %s\n\t%s" %
                                (visit_node, e))

    def write_visit(self, visit_node, indent_level=0):
        try:
            if visit_node.node_repr is not None:
                self.output.write(self.indent_str * indent_level)
                self.output.write(visit_node.node_repr)
                self.output.write('\n')
        except AttributeError:
            raise TreeWalkError("can't write visit_node: %s" % visit_node)

        for cn in visit_node.child_nodes:
            self.write_visit(cn, indent_level + 1)

    def build_text(self, ast_node):
        method_name = 'visitAST%s' % ast_node.__class__.__name__
        method = getattr(self, method_name, self.visitDefault)
        return method(ast_node)

    def visitASTTemplateNode(self, node):
        module_code = self.visitDefault(node)[0]
        if node.import_nodes:
            for n in node.import_nodes:
                module_code.extend(self.build_text(n))
        if node.from_nodes:
            for n in node.from_nodes:
                module_code.extend(self.build_text(n))
        if node.extends_nodes:
            for n in node.extends_nodes:
                module_code.extend(self.build_text(n))

        for n in node.attr_nodes:
            module_code.extend(self.build_text(n))

        for n in node.child_nodes:
            module_code.extend(self.build_text(n))

        # if we aren't extending a template, build out the main function
        if not node.extends_nodes and not node.library:
            module_code.extend(self.build_text(node.main_function))

        return [module_code]

    def visitASTExtendsNode(self, node):
        v = self.visitDefault(node)[0]
        for n in node.module_name_list:
            v.extend(self.build_text(n))
        return [v]

    visitASTImportNode = visitASTExtendsNode
    visitASTFromNode = visitASTExtendsNode
    visitASTAbsoluteExtendsNode = visitASTExtendsNode

    def visitASTCallFunctionNode(self, node):
        v = self.visitDefault(node)[0]
        v.append(VisitNode('expression', self.build_text(node.expression)))
        if node.arg_list:
            v.extend(self.build_text(node.arg_list))
        return [v]

    def visitASTCacheNode(self, node):
        v = self.visitDefault(node)[0]
        v.append(VisitNode('expression', self.build_text(node.expression)))
        return [v]

    def visitASTForNode(self, node):
        visit_node = self.visitDefault(node)[0]
        target_list = self.build_text(node.target_list)
        expression_list = self.build_text(node.expression_list)
        visit_node.extend(target_list)
        visit_node.extend(expression_list)
        for n in node.child_nodes:
            visit_node.extend(self.build_text(n))
        return [visit_node]

    def visitASTIfNode(self, node):
        if_node = self.visitDefault(node)[0]
        test_expression = VisitNode('test_expression',
                                    self.build_text(node.test_expression))
        if_node.append(test_expression)
        #if_node.append(VisitNode(str(node.scope)))
        for n in node.child_nodes:
            if_node.extend(self.build_text(n))
        visit_nodes = [if_node]
        if node.else_.child_nodes:
            else_visit_node = VisitNode('else')
            for n in node.else_.child_nodes:
                else_visit_node.extend(self.build_text(n))
            visit_nodes.append(else_visit_node)
        return visit_nodes

    def visitASTAttributeNode(self, node):
        v = self.visitDefault(node)[0]
        v.extend(self.build_text(node.default))
        return [v]

    def visitASTEchoNode(self, node):
        v = self.visitDefault(node)[0]
        for n in [node.test_expression, node.true_expression,
                  node.false_expression]:
            #print "visitASTParameterListNode:", n, text
            v.append(VisitNode(str(n)))
        return [v]

    def visitASTParameterListNode(self, node):
        v = self.visitDefault(node)[0]
        for n in node.child_nodes:
            #print "visitASTParameterListNode:", n, text
            v.append(VisitNode(str(n)))
        return [v]

    visitASTArgListNode = visitASTParameterListNode
    visitASTTargetListNode = visitASTParameterListNode
    visitASTExpressionListNode = visitASTParameterListNode
    visitASTListLiteralNode = visitASTParameterListNode
    visitASTTupleLiteralNode = visitASTParameterListNode

    def visitASTDictLiteralNode(self, node):
        v = self.visitDefault(node)[0]
        for n in node.child_nodes:
            key_expression, value_expression = n
            v.extend(self.build_text(key_expression))
            v.extend(self.build_text(value_expression))
        return [v]

    def visitASTParameterNode(self, node):
        if node.default:
            return [VisitNode('%s=%s' %
                              (node.name, ' '.join([self.generate_text(
                                  x) for x in self.build_text(node.default)])))]
        else:
            return [VisitNode('%s' % node.name)]

    def visitASTGetUDNNode(self, node):
        v = self.visitDefault(node)[0]
        v.extend(self.build_text(node.expression))
        return [v]

    visitASTGetAttrNode = visitASTGetUDNNode
    visitASTReturnNode = visitASTGetUDNNode
    visitASTPlaceholderSubstitutionNode = visitASTGetUDNNode
    visitASTBufferWrite = visitASTGetUDNNode

    def visitASTFilterNode(self, node):
        v = VisitNode('%s %s %s' %
                      (node.__class__.__name__, node.name, hash(node)))
        v.extend(self.build_text(node.expression))
        return [v]

    def visitASTSliceNode(self, node):
        v = self.visitDefault(node)[0]
        v.append(VisitNode('expression', self.build_text(node.expression)))
        v.append(VisitNode('slice', self.build_text(node.slice_expression)))
        return [v]

    def visitASTBinOpExpressionNode(self, node):
        v = self.visitDefault(node)[0]
        v.extend(self.build_text(node.left))
        v.append(VisitNode(node.operator))
        v.extend(self.build_text(node.right))
        return [v]

    visitASTBinOpNode = visitASTBinOpExpressionNode

    visitASTAssignNode = visitASTBinOpNode

    def visitASTUnaryOpNode(self, node):
        v = self.visitDefault(node)[0]
        v.append(VisitNode(node.operator))
        v.extend(self.build_text(node.expression))
        return [v]

    def visitASTFunctionNode(self, node):
        v = self.visitDefault(node)[0]
        if node.parameter_list:
            v.extend(self.build_text(node.parameter_list))
        #v.append(VisitNode(str(node.scope)))

        for n in node.child_nodes:
            v.extend(self.build_text(n))

        return [v]

    def visitASTDefNode(self, node):
        v = self.visitDefault(node)[0]
        if node.parameter_list:
            v.extend(self.build_text(node.parameter_list))

        for n in node.child_nodes:
            v.extend(self.build_text(n))

        return [v]

    visitASTBlockNode = visitASTDefNode

    def visitASTFragmentNode(self, node):
        v = self.visitDefault(node)[0]
        for n in node.child_nodes:
            v.extend(self.build_text(n))

        return [v]

    visitASTStripLinesNode = visitASTFragmentNode

    def visitASTDoNode(self, node):
        v = self.visitDefault(node)[0]
        v.extend(self.build_text(node.expression))
        return [v]

    def visitASTLiteralNode(self, node):
        return [VisitNode("%s '%r'" % (node.__class__.__name__, node.value))]

    visitASTTextNode = visitASTLiteralNode
    visitASTWhitespaceNode = visitASTLiteralNode
    visitASTOptionalWhitespaceNode = visitASTLiteralNode

    def visitDefault(self, node):
        return [VisitNode('%s %s' % (node.__class__.__name__, node.name))]

    def visitASTMacroNode(self, node):
        return [VisitNode('%s %s' % (node.__class__.__name__, node.value))]



# Copyright 2008 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# a few helpful filter functions

import functools
import types
from spitfire import runtime
from spitfire.runtime import udn


# decorate a function object so the default filter will not be applied to the
# value of a placeholder. this is handy when building functions that will
# create data that could be double-escaped and you don't wnat to constantly
# inform spitfire to us raw mode.
def skip_filter(function):
    if isinstance(function, types.BuiltinFunctionType):
        # Built-in functions don't allow abitrary attributes, so we use a
        # wrapper function that just passes through
        @functools.wraps(function)
        def skip_filter_wrapper(*args, **kwargs):
            return function(*args, **kwargs)
        skip_filter_wrapper.skip_filter = True
        return skip_filter_wrapper

    function.skip_filter = True
    return function


def passthrough_filter(value):
    return value


@skip_filter
def escape_html(value, quote=True):
    """Replace special characters '&', '<' and '>' by SGML entities."""
    value = simple_str_filter(value)
    if isinstance(value, str):
        value = value.replace("&", "&amp;")  # Must be done first!
        value = value.replace("<", "&lt;")
        value = value.replace(">", "&gt;")
        if quote:
            value = value.replace('"', "&quot;")
    return value


# deprecated
def safe_values(value):
    """Deprecated - use simple_str_filter instead."""
    if isinstance(value, (str, int, float,
                          runtime.UndefinedPlaceholder)):
        return value
    else:
        return ''


def simple_str_filter(value):
    """Return a string if the input type is something primitive."""
    if isinstance(value, (str, int, float,
                          runtime.UndefinedPlaceholder)):
        # fixme: why do force this conversion here?
        # do we want to be unicode or str?
        return str(value)
    else:
        return ''


# test function for function registry - don't use
@skip_filter
def escape_html_function(value):
    return escape_html(value)



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import logging
import types
import os.path
import re
import sys


from spitfire.compiler import options
from spitfire.compiler import parser
from spitfire.compiler import scanner
from spitfire import runtime

valid_identfier = re.compile('[_a-z]\w*', re.IGNORECASE)


def filename2classname(filename):
    classname = os.path.splitext(os.path.basename(filename))[0].replace('-',
                                                                        '_')
    if not valid_identfier.match(classname):
        raise SyntaxError('filename "%s" must yield valid python identifier: %s'
                          % (filename, classname))
    return classname


def filename2modulename(filename):
    names = [filename2classname(p) for p in filename.split(os.path.sep)]
    return '.'.join(names)


# @return abstract syntax tree rooted on a ast.TemplateNode
def parse(src_text, rule='goal'):
    spt_parser = parser.SpitfireParser(scanner.SpitfireScanner(src_text))
    return parser.wrap_error_reporter(spt_parser, rule)


def parse_file(filename, xspt_mode=False):
    template_node = parse_template(read_template_file(filename), xspt_mode)
    template_node.source_path = filename
    return template_node


def parse_template(src_text, xspt_mode=False):
    if xspt_mode:
        # Note: The compiler module is imported here to avoid a circular
        # dependency.
        from spitfire.compiler import xhtml2ast
        xspt_parser = xhtml2ast.XHTML2AST()
        return xspt_parser.parse(src_text)
    else:
        return parse(src_text)


def read_template_file(filename):
    f = open(filename, 'r')
    try:
        return f.read()
    finally:
        f.close()


def read_function_registry(filename):
    f = open(filename)
    lines = f.readlines()
    f.close()
    new_format = [l for l in lines
                  if not l.startswith('#') and l.find(',') > -1]
    function_registry = {}
    for line in lines:
        line = line.strip()
        if not line:
            continue
        if line.startswith('#'):
            continue
        if new_format:
            alias, rest = line.split('=')
            decorators = rest.split(',')
            fq_name = decorators.pop(0).strip()
            function_registry[alias.strip()] = fq_name, decorators
        else:
            alias, fq_name = line.split('=')
            fq_name = fq_name.strip()
            try:
                method = runtime.import_module_symbol(fq_name)
            except ImportError:
                logging.warning('unable to import function registry symbol %s',
                                fq_name)
                method = None
            function_registry[alias.strip()] = fq_name, method
    return new_format, function_registry


# compile a text file into a template object
# this won't recursively import templates, it's just a convenience in the case
# where you need to create a fresh object directly from raw template file
def load_template_file(filename,
                       module_name=None,
                       analyzer_options=options.default_options,
                       compiler_options=None,
                       xspt_mode=False):
    # Note: The compiler module is imported here to avoid a circular dependency.
    from spitfire.compiler import compiler
    spt_compiler = compiler.Compiler(analyzer_options=analyzer_options,
                                     xspt_mode=xspt_mode)
    if compiler_options:
        for k, v in compiler_options.items():
            setattr(spt_compiler, k, v)
    class_name = filename2classname(filename)
    if not module_name:
        module_name = class_name

    src_code = spt_compiler.compile_file(filename)
    module = load_module_from_src(src_code, filename, module_name)
    return getattr(module, class_name)


def load_template(template_src,
                  template_name,
                  analyzer_options=options.default_options,
                  compiler_options=None):
    class_name = filename2classname(template_name)
    filename = '<+' + class_name + '>'
    module_name = class_name
    # Note: The compiler module is imported here to avoid a circular dependency
    from spitfire.compiler import compiler
    spt_compiler = compiler.Compiler(analyzer_options=analyzer_options)
    if compiler_options:
        for k, v in compiler_options.items():
            setattr(spt_compiler, k, v)
    src_code = spt_compiler.compile_template(template_src, class_name)
    module = load_module_from_src(src_code, filename, module_name)
    return getattr(module, class_name)


# a helper method to import a template without having to save it to disk
def load_module_from_src(src_code, filename, module_name):
    module = types.ModuleType(module_name)
    sys.modules[module_name] = module

    bytecode = compile(src_code, filename, 'exec')
    exec(bytecode, module.__dict__)
    return module


# convert and extends path to a file path
def extends2path(class_extend):
    return class_extend.replace('.', '/') + '.spt'



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import logging
import optparse
import os.path
import sys

import pickle as pickle


def run_template(class_object):
    option_parser = optparse.OptionParser()
    option_parser.add_option('-f',
                             dest='filename',
                             help='data file for template search list')
    (tmpl_options, tmpl_args) = option_parser.parse_args()

    if tmpl_options.filename:
        data = [load_search_list(tmpl_options.filename)]
    else:
        data = []
    template = class_object(search_list=data)
    sys.stdout.write(template.main())


def load_search_list(filename):
    f = open(filename)
    raw_data = f.read()
    ext = os.path.splitext(filename)[-1]
    if ext == '.pkl':
        data = pickle.loads(raw_data)
    else:
        try:
            data = eval(raw_data)
        except Exception as e:
            logging.error('load_search_list\n%s', raw_data)
            raise
    return data



# Copyright 2007 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

# -*- encoding: utf-8 -*-
# Template language benchmarks
#
# Objective: Generate a 1000x10 HTML table as fast as possible.
#
# Author: Jonas Borgström <jonas@edgewall.com>

import cgi
import sys
import timeit
import io


try:
    import genshi
    from genshi.builder import tag
    from genshi.template import MarkupTemplate
    from genshi.template import TextTemplate as NewTextTemplate
except ImportError:
    MarkupTemplate = None
    NewTextTemplate = None
    genshi = None

try:
    from elementtree import ElementTree as et
except ImportError:
    et = None

try:
    import cElementTree as cet
except ImportError:
    cet = None

try:
    import neo_cgi, neo_cs, neo_util
except ImportError:
    neo_cgi = None

try:
    import kid
except ImportError:
    kid = None

try:
    from django.conf import settings
    settings.configure()
    from django.template import Context as DjangoContext
    from django.template import Template as DjangoTemplate
except ImportError:
    DjangoContext = DjangoTemplate = None

try:
    from mako.template import Template as MakoTemplate
except ImportError:
    MakoTemplate = None

try:
    import spitfire as SpitfireTemplate
except ImportError:
    SpitfireTemplate = None

try:
    from Cheetah import Template as CheetahTemplate
except ImportError:
    CheetahTemplate = None

table = [dict(a=1,b=2,c=3,d=4,e=5,f=6,g=7,h=8,i=9,j=10)
          for x in range(1000)]

if MarkupTemplate:
    genshi_tmpl = MarkupTemplate("""
<table xmlns:py="http://genshi.edgewall.org/">
<tr py:for="row in table">
<td py:for="c in row.values()" py:content="c"/>
</tr>
</table>
""")

    genshi_tmpl2 = MarkupTemplate("""
<table xmlns:py="http://genshi.edgewall.org/">$table</table>
""")

    genshi_text_tmpl = NewTextTemplate("""
<table>
{% for row in table %}<tr>
{% for c in row.values() %}<td>{c}</td>{% end %}
</tr>{% end %}
</table>
""")

if MakoTemplate:
    mako_tmpl = MakoTemplate("""
<table>
  % for row in table:
    <tr>
      % for col in row.values():
        <td>${ col | h  }</td>
      % endfor
    </tr>
  % endfor
</table>
""")
    def test_mako():
        """Mako Template"""
        data = mako_tmpl.render(table=table)
        #print "mako", len(data)

if SpitfireTemplate:
    import spitfire.compiler.analyzer
    import spitfire.compiler.util
    spitfire_src = """<table xmlns:py="http://spitfire/">
#for $row in $table
<tr>
#for $column in $row.values()
<td>$column</td>
#end for
</tr>
#end for
</table>
"""
    enable_filters = False
    spitfire_tmpl = spitfire.compiler.util.load_template(
        spitfire_src, 'spitfire_tmpl')

    spitfire_tmpl_o1 = spitfire.compiler.util.load_template(
        spitfire_src, 'spitfire_tmpl_o1', spitfire.compiler.options.o1_options,
        {'enable_filters':enable_filters})

    spitfire_tmpl_o2 = spitfire.compiler.util.load_template(
        spitfire_src, 'spitfire_tmpl_o2', spitfire.compiler.options.o2_options,
        {'enable_filters':enable_filters})

    spitfire_tmpl_o3 = spitfire.compiler.util.load_template(
        spitfire_src, 'spitfire_tmpl_o3', spitfire.compiler.options.o3_options,
        {'enable_filters':enable_filters})


    def test_spitfire():
        """Spitfire template"""
        data = spitfire_tmpl(search_list=[{'table':table}]).main()
        #print "spitfire", len(data)

    def test_spitfire_o1():
        """Spitfire template -O1"""
        data = spitfire_tmpl_o1(search_list=[{'table':table}]).main()
        #print "spitfire -O1", len(data)

    def test_spitfire_o2():
        """Spitfire template -O2"""
        data = spitfire_tmpl_o2(search_list=[{'table':table}]).main()

    def test_spitfire_o3():
        """Spitfire template -O3"""
        data = spitfire_tmpl_o3(search_list=[{'table':table}]).main()
        #print "spitfire -O3", len(data)

if CheetahTemplate:
    cheetah_src = """<table>
#for $row in $table
<tr>
#for $column in $row.values()
<td>$column</td>
#end for
</tr>
#end for
</table>"""
    pre = set([k for k, v in sys.modules.items() if v])
    cheetah_template = CheetahTemplate.Template(cheetah_src, searchList=[{'table':table}])
    # force compile
    post = set([k for k, v in sys.modules.items() if v])
    #print post - pre

    #print type(cheetah_template)
    cheetah_template.respond()
    cheetah_template = type(cheetah_template)

    def test_cheetah():
        """Cheetah template"""
        data = cheetah_template(searchList=[{'table':table}]).respond()
        #print "cheetah", len(data)

if genshi:
    def test_genshi():
        """Genshi template"""
        stream = genshi_tmpl.generate(table=table)
        data = stream.render('html', strip_whitespace=False)
        #print "genshi", len(data)

    def disabled_test_genshi_text():
        """Genshi text template"""
        stream = genshi_text_tmpl.generate(table=table)
        print("test_genshi_text", stream)
        data = stream.render('text')
        print("test_genshi_text", 'data', stream)

    def test_genshi_builder():
        """Genshi template + tag builder"""
        stream = tag.TABLE([
            tag.tr([tag.td(c) for c in list(row.values())])
            for row in table
        ]).generate()
        stream = genshi_tmpl2.generate(table=stream)
        stream.render('html', strip_whitespace=False)

    def test_builder():
        """Genshi tag builder"""
        stream = tag.TABLE([
            tag.tr([
                tag.td(c) for c in list(row.values())
            ])
            for row in table
        ]).generate()
        stream.render('html', strip_whitespace=False)

if kid:
    kid_tmpl = kid.Template("""
    <table xmlns:py="http://purl.org/kid/ns#">
    <tr py:for="row in table">
    <td py:for="c in row.values()" py:content="c"/>
    </tr>
    </table>
    """)

    kid_tmpl2 = kid.Template("""
    <html xmlns:py="http://purl.org/kid/ns#">$table</html>
    """)

    def test_kid():
        """Kid template"""
        kid_tmpl.table = table
        kid_tmpl.serialize(output='html')

    if cet:
        def test_kid_et():
            """Kid template + cElementTree"""
            _table = cet.Element('table')
            for row in table:
                td = cet.SubElement(_table, 'tr')
                for c in list(row.values()):
                    cet.SubElement(td, 'td').text=str(c)
            kid_tmpl2.table = _table
            kid_tmpl2.serialize(output='html')

if et:
    def test_et():
        """ElementTree"""
        _table = et.Element('table')
        for row in table:
            tr = et.SubElement(_table, 'tr')
            for c in list(row.values()):
                et.SubElement(tr, 'td').text=str(c)
        et.tostring(_table)

if cet:
    def test_cet():
        """cElementTree"""
        _table = cet.Element('table')
        for row in table:
            tr = cet.SubElement(_table, 'tr')
            for c in list(row.values()):
                cet.SubElement(tr, 'td').text=str(c)
        cet.tostring(_table)

if neo_cgi:
    def test_clearsilver():
        """ClearSilver"""
        hdf = neo_util.HDF()
        for i, row in enumerate(table):
            for j, c in enumerate(row.values()):
                hdf.setValue("rows.%d.cell.%d" % (i, j), cgi.escape(str(c)))

        cs = neo_cs.CS(hdf)
        cs.parseStr("""
<table><?cs
  each:row=rows
?><tr><?cs each:c=row.cell
  ?><td><?cs var:c ?></td><?cs /each
?></tr><?cs /each?>
</table>""")
        cs.render()

def test_python_cstringio():
    """cStringIO"""
    buffer = io.StringIO()
    write = buffer.write
    write('<table>\n')
    for row in table:
        write('<tr>\n')
        for col in row.values():
            write('<td>\n')
            write('%s' % col)
            write('\n</td>\n')
        write('</tr>\n')
    write('</table>')
    return buffer.getvalue()

def test_python_stringio():
    """StringIO"""
    buffer = io.StringIO()
    write = buffer.write

    write('<table>\n')
    for row in table:
        write('<tr>\n')
        for col in row.values():
            write('<td>\n')
            write('%s' % col)
            write('\n</td>\n')
        write('</tr>\n')
    write('</table>')
    return buffer.getvalue()

def test_python_array():
    """list concat"""
    buffer = []
    write = buffer.append

    write('<table>\n')
    for row in table:
        write('<tr>\n')
        for col in row.values():
            write('<td>\n')
            write('%s' % col)
            write('\n</td>\n')
        write('</tr>\n')
    write('</table>')
    return ''.join(buffer)


def run(which=None, number=10):
    tests = ['test_builder', 'test_genshi',
             'test_genshi_text',
             'test_genshi_builder', 'test_mako', 'test_kid', 'test_kid_et',
             'test_et', 'test_cet', 'test_clearsilver', 'test_django',
             'test_cheetah',
             'test_spitfire', 'test_spitfire_o1',
             'test_spitfire_o2', 'test_spitfire_o3',
             'test_python_stringio', 'test_python_cstringio', 'test_python_array'
             ]

    if which:
        tests = [n for n in tests if n[5:] in which]

    for test in [t for t in tests if hasattr(sys.modules[__name__], t)]:
        t = timeit.Timer(setup='from __main__ import %s;' % test,
                         stmt='%s()' % test)
        time = t.timeit(number=number) / number

        if time < 0.00001:
            result = '   (not installed?)'
        else:
            result = '%16.2f ms' % (1000 * time)
        print('%-35s %s' % (getattr(sys.modules[__name__], test).__doc__, result))


if __name__ == '__main__':
    which = [arg for arg in sys.argv[1:] if arg[0] != '-']

    if '-p' in sys.argv:
        import hotshot, hotshot.stats
        prof = hotshot.Profile("template.prof")
        benchtime = prof.runcall(run, which, number=1)
        stats = hotshot.stats.load("template.prof")
        stats.strip_dirs()
        stats.sort_stats('time', 'calls')
        stats.print_stats()
    else:
        run(which, 1)



# Copyright 2016 The Spitfire Authors. All Rights Reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file.

import copy
import optparse
import string
import io
import io
import sys
import timeit

try:
    import spitfire
    import spitfire.compiler.util
    import spitfire.compiler.options
except ImportError:
    spitfire = None

try:
    import Cheetah
    import Cheetah.Template
except ImportError:
    Cheetah = None

try:
    import django
    import django.conf
    import django.template
except ImportError:
    django = None

try:
    import jinja2
except ImportError:
    jinja2 = None

try:
    import mako
    import mako.template
except ImportError:
    mako = None


TABLE_DATA = [
    dict(a=1,
         b=2,
         c=3,
         d=4,
         e=5,
         f=6,
         g=7,
         h=8,
         i=9,
         j=10) for x in range(1000)
]


def get_spitfire_tests():
    if not spitfire:
        return []

    tmpl_src = """
        <table>
            #for $row in $table
                <tr>
                    #for $column in $row.values()
                        <td>$column</td>
                    #end for
                </tr>
            #end for
        </table>
    """

    tmpl_search_list = [{'table': TABLE_DATA}]

    default_opts = spitfire.compiler.options.default_options
    o1_opts = spitfire.compiler.options.o1_options
    o2_opts = spitfire.compiler.options.o2_options
    o3_opts = spitfire.compiler.options.o3_options

    def _spitfire_baked_opts(o):
        o = copy.copy(o)
        o.baked_mode = True
        o.generate_unicode = False
        return o

    baked_opts = _spitfire_baked_opts(default_opts)
    baked_o1_opts = _spitfire_baked_opts(o1_opts)
    baked_o2_opts = _spitfire_baked_opts(o2_opts)
    baked_o3_opts = _spitfire_baked_opts(o3_opts)

    tmpl = spitfire.compiler.util.load_template(tmpl_src,
                                                'tmpl',
                                                analyzer_options=default_opts)

    tmpl_o1 = spitfire.compiler.util.load_template(tmpl_src,
                                                   'tmpl_o1',
                                                   analyzer_options=o1_opts)

    tmpl_o2 = spitfire.compiler.util.load_template(tmpl_src,
                                                   'tmpl_o2',
                                                   analyzer_options=o2_opts)

    tmpl_o3 = spitfire.compiler.util.load_template(tmpl_src,
                                                   'tmpl_o3',
                                                   analyzer_options=o3_opts)

    tmpl_baked = spitfire.compiler.util.load_template(
        tmpl_src,
        'tmpl_baked',
        analyzer_options=baked_opts)

    tmpl_baked_o1 = spitfire.compiler.util.load_template(
        tmpl_src,
        'tmpl_baked_o1',
        analyzer_options=baked_o1_opts)

    tmpl_baked_o2 = spitfire.compiler.util.load_template(
        tmpl_src,
        'tmpl_baked_o2',
        analyzer_options=baked_o2_opts)

    tmpl_baked_o3 = spitfire.compiler.util.load_template(
        tmpl_src,
        'tmpl_baked_o3',
        analyzer_options=baked_o3_opts)

    tmpl_unfiltered = spitfire.compiler.util.load_template(
        tmpl_src,
        'tmpl_unfiltered',
        analyzer_options=default_opts,
        compiler_options={'enable_filters': False})

    tmpl_unfiltered_o1 = spitfire.compiler.util.load_template(
        tmpl_src,
        'tmpl_unfiltered_o1',
        analyzer_options=o1_opts,
        compiler_options={'enable_filters': False})

    tmpl_unfiltered_o2 = spitfire.compiler.util.load_template(
        tmpl_src,
        'tmpl_unfiltered_o2',
        analyzer_options=o2_opts,
        compiler_options={'enable_filters': False})

    tmpl_unfiltered_o3 = spitfire.compiler.util.load_template(
        tmpl_src,
        'tmpl_unfiltered_o3',
        analyzer_options=o3_opts,
        compiler_options={'enable_filters': False})

    def test_spitfire():
        """Spitfire template"""

    def test_spitfire_o1():
        """Spitfire template -O1"""

    def test_spitfire_o2():
        """Spitfire template -O2"""
        tmpl_o2(search_list=tmpl_search_list).main()

    def test_spitfire_o3():
        """Spitfire template -O3"""
        tmpl_o3(search_list=tmpl_search_list).main()

    def test_spitfire_baked():
        """Spitfire template baked"""
        tmpl_baked(search_list=tmpl_search_list).main()

    def test_spitfire_baked_o1():
        """Spitfire template baked -O1"""
        tmpl_baked_o2(search_list=tmpl_search_list).main()

    def test_spitfire_baked_o2():
        """Spitfire template baked -O2"""
        tmpl_baked_o2(search_list=tmpl_search_list).main()

    def test_spitfire_baked_o3():
        """Spitfire template baked -O3"""
        tmpl_baked_o3(search_list=tmpl_search_list).main()

    def test_spitfire_unfiltered():
        """Spitfire template unfiltered"""
        tmpl_unfiltered(search_list=tmpl_search_list).main()

    def test_spitfire_unfiltered_o1():
        """Spitfire template unfiltered -O1"""
        tmpl_unfiltered_o2(search_list=tmpl_search_list).main()

    def test_spitfire_unfiltered_o2():
        """Spitfire template unfiltered -O2"""
        tmpl_unfiltered_o2(search_list=tmpl_search_list).main()

    def test_spitfire_unfiltered_o3():
        """Spitfire template unfiltered -O3"""
        tmpl_unfiltered_o3(search_list=tmpl_search_list).main()

    return [
        test_spitfire,
        test_spitfire_o1,
        test_spitfire_o2,
        test_spitfire_o3,
        test_spitfire_baked,
        test_spitfire_baked_o1,
        test_spitfire_baked_o2,
        test_spitfire_baked_o3,
        test_spitfire_unfiltered,
        test_spitfire_unfiltered_o1,
        test_spitfire_unfiltered_o2,
        test_spitfire_unfiltered_o3,
    ]


def get_python_tests():
    tmpl_table = string.Template('<table>\n$table\n</table>\n')
    tmpl_row = string.Template('<tr>\n$row\n</tr>\n')
    tmpl_column = string.Template('<td>$column</td>\n')

    def _buffer_fn(write, table):
        write('<table>\n')
        for row in table:
            write('<tr>\n')
            for column in row.values():
                write('<td>')
                write('%s' % column)
                write('</td>\n')
            write('</tr>\n')
        write('</table>\n')

    def test_python_template():
        """Python string template"""
        rows = ''
        for row in TABLE_DATA:
            columns = ''
            for column in row.values():
                columns = columns + tmpl_column.substitute(column=column)
            rows = rows + tmpl_row.substitute(row=columns)
        return tmpl_table.substitute(table=rows)

    def test_python_stringio():
        """Python StringIO buffer"""
        buffer = io.StringIO()
        _buffer_fn(buffer.write, TABLE_DATA)
        return buffer.getvalue()

    def test_python_cstringio():
        """Python cStringIO buffer"""
        buffer = io.StringIO()
        _buffer_fn(buffer.write, TABLE_DATA)
        return buffer.getvalue()

    def test_python_list():
        """Python list concatenation"""
        buffer = []
        _buffer_fn(buffer.append, TABLE_DATA)
        return ''.join(buffer)

    return [
        test_python_template,
        test_python_stringio,
        test_python_cstringio,
        test_python_list,
    ]


def get_cheetah_tests():
    if not Cheetah:
        return []

    tmpl_src = """
        <table>
            #for $row in $table
                <tr>
                    #for $column in $row.values()
                        <td>$column</td>
                    #end for
                </tr>
            #end for
        </table>
    """

    tmpl_search_list = [{'table': TABLE_DATA}]

    tmpl = Cheetah.Template.Template(tmpl_src, searchList=tmpl_search_list)

    def test_cheetah():
        """Cheetah template"""
        tmpl.respond()

    return [
        test_cheetah,
    ]


def get_django_tests():
    if not django:
        return []

    django.conf.settings.configure()
    django.setup()

    tmpl_src = """
        <table>
            {% for row in table %}
                <tr>
                    {% for column in row.values %}
                        <td>{{ column }}</td>
                    {% endfor %}
                </tr>
            {% endfor %}
        </table>
    """
    tmpl_autoescaped_src = ('{% autoescape on %}' +
                            tmpl_src +
                            '{% endautoescape %}')

    tmpl = django.template.Template(tmpl_src)
    tmpl_autoescaped = django.template.Template(tmpl_autoescaped_src)

    tmpl_context = django.template.Context({'table': TABLE_DATA})

    def test_django():
        """Django template"""
        tmpl.render(tmpl_context)

    def test_django_autoescaped():
        """Django template autoescaped"""
        tmpl_autoescaped.render(tmpl_context)

    return [
        test_django,
        test_django_autoescaped,
    ]


def get_jinja2_tests():
    if not jinja2:
        return []

    tmpl_src = """
        <table>
            {% for row in table %}
                <tr>
                    {% for column in row.values() %}
                        <td>{{ column }}</td>
                    {% endfor %}
                </tr>
            {% endfor %}
        </table>
    """

    tmpl = jinja2.Template(tmpl_src)
    tmpl_autoescaped = jinja2.Template(tmpl_src, autoescape=True)

    def test_jinja2():
        """Jinja2 template"""
        tmpl.render(table=TABLE_DATA)

    def test_jinja2_autoescaped():
        """Jinja2 template autoescaped"""
        tmpl_autoescaped.render(table=TABLE_DATA)

    return [
        test_jinja2,
        test_jinja2_autoescaped,
    ]


def get_mako_tests():
    if not mako:
        return []

    tmpl_src = """
        <table>
            % for row in table:
                <tr>
                    % for column in row.values():
                        <td>${column}</td>
                    % endfor
                </tr>
            % endfor
        </table>
    """

    tmpl = mako.template.Template(tmpl_src)
    tmpl_autoescaped = mako.template.Template(tmpl_src, default_filters=['h'])

    def test_mako():
        """Mako template"""
        tmpl.render(table=TABLE_DATA)

    def test_mako_autoescaped():
        """Mako template autoescaped"""
        tmpl_autoescaped.render(table=TABLE_DATA)

    return [
        test_mako,
        test_mako_autoescaped,
    ]


def time_test(test, number):
    # Put the test in the global scope for timeit.
    name = 'timeit_%s' % test.__name__
    globals()[name] = test
    # Time the test.
    timer = timeit.Timer(setup='from __main__ import %s;' % name,
                         stmt='%s()' % name)
    time = timer.timeit(number=number) / number
    if time < 0.00001:
        result = '   (not installed?)'
    else:
        result = '%16.2f ms' % (1000 * time)
    print('%-35s %s' % (test.__doc__, result))


def run_tests(which=None, number=100, compare=False):
    if number > 100:
        print('Running benchmarks %d times each...' % number)
        print()
    if compare:
        groups = ['cheetah', 'django', 'jinja2', 'mako', 'python', 'spitfire']
    else:
        groups = ['spitfire']
    # Built the full list of eligible tests.
    tests = []
    missing_engines = []
    for g in groups:
        test_list_fn = 'get_%s_tests' % g
        test = globals()[test_list_fn]()
        if test:
          tests.extend(test)
        else:
          missing_engines.append(g)
    # Optionally filter by a set of matching test name (sub)strings.
    if which:
        which_tests = []
        for t in tests:
            for w in which:
                if w.lower() in t.__name__.lower():
                    which_tests.append(t)
        tests = which_tests
    # Report any missing template engines.
    if missing_engines:
        sys.stderr.write(
            'The following template engines are not installed and will be '
            'skipped in the benchmark: %r\n' % missing_engines)
    # Run the tests.
    for t in tests:
        time_test(t, number)


def profile_tests(which=None):
    print('Profiling...')
    print()
    import hotshot, hotshot.stats
    profile_data = 'template.prof'
    profile = hotshot.Profile(profile_data)
    profile.runcall(run_tests, which=which, number=1, compare=False)
    stats = hotshot.stats.load(profile_data)
    stats.strip_dirs()
    stats.sort_stats('time', 'calls')
    print()
    stats.print_stats()
    print('Profile data written to %s' % profile_data)


def main():
    option_parser = optparse.OptionParser()
    option_parser.add_option('-n', '--number', type='int', default=100)
    option_parser.add_option('-c',
                             '--compare',
                             action='store_true',
                             default=False)
    option_parser.add_option('-p',
                             '--profile',
                             action='store_true',
                             default=False)
    (options, args) = option_parser.parse_args()

    if options.profile:
        profile_tests(which=args)
    else:
        run_tests(which=args, number=options.number, compare=options.compare)


if __name__ == '__main__':
    main()
